import nengo
import numpy as np

from nengo_spa.algebras.base import AbstractAlgebra
from nengo_spa.networks.circularconvolution import CircularConvolution


class HrrAlgebra(AbstractAlgebra):
    r"""Holographic Reduced Representations (HRRs) algebra.

    Uses element-wise addition for superposition, circular convolution for
    binding with an approximate inverse.

    The circular convolution :math:`c` of vectors :math:`a` and :math:`b`
    is given by

    .. math:: c[i] = \sum_j a[j] b[i - j]

    where negative indices on :math:`b` wrap around to the end of the vector.

    This computation can also be done in the Fourier domain,

    .. math:: c = DFT^{-1} ( DFT(a) \odot DFT(b) )

    where :math:`DFT` is the Discrete Fourier Transform operator, and
    :math:`DFT^{-1}` is its inverse.

    Circular convolution as a binding operation is associative, commutative,
    distributive.

    More information on circular convolution as a binding operation can be
    found in [plate2003]_.

    .. [plate2003] Plate, Tony A. Holographic Reduced Representation:
       Distributed Representation for Cognitive Structures. Stanford, CA: CSLI
       Publications, 2003.
    """

    _instance = None

    def __new__(cls):
        if type(cls._instance) is not cls:
            cls._instance = super(HrrAlgebra, cls).__new__(cls)
        return cls._instance

    def is_valid_dimensionality(self, d):
        """Checks whether *d* is a valid vector dimensionality.

        For circular convolution all positive numbers are valid
        dimensionalities.

        Parameters
        ----------
        d : int
            Dimensionality

        Returns
        -------
        bool
            *True*, if *d* is a valid vector dimensionality for the use with
            the algebra.
        """
        return d > 0

    def make_unitary(self, v):
        fft_val = np.fft.fft(v)
        fft_imag = fft_val.imag
        fft_real = fft_val.real
        fft_norms = np.sqrt(fft_imag ** 2 + fft_real ** 2)
        invalid = fft_norms <= 0.0
        fft_val[invalid] = 1.0
        fft_norms[invalid] = 1.0
        fft_unit = fft_val / fft_norms
        return np.array((np.fft.ifft(fft_unit, n=len(v))).real)

    def superpose(self, a, b):
        return a + b

    def bind(self, a, b):
        n = len(a)
        if len(b) != n:
            raise ValueError("Inputs must have same length.")
        return np.fft.irfft(np.fft.rfft(a) * np.fft.rfft(b), n=n)

    def invert(self, v):
        return v[-np.arange(len(v))]

    def get_binding_matrix(self, v, swap_inputs=False):
        D = len(v)
        T = []
        for i in range(D):
            T.append([v[(i - j) % D] for j in range(D)])
        return np.array(T)

    def get_inversion_matrix(self, d):
        return np.eye(d)[-np.arange(d)]

    def implement_superposition(self, n_neurons_per_d, d, n):
        node = nengo.Node(size_in=d)
        return node, n * (node,), node

    def implement_binding(self, n_neurons_per_d, d, unbind_left, unbind_right):
        net = CircularConvolution(n_neurons_per_d, d, unbind_left, unbind_right)
        return net, (net.input_a, net.input_b), net.output

    def absorbing_element(self, d):
        r"""Return the standard absorbing element of dimensionality *d*.

        An absorbing element will produce a scaled version of itself when bound
        to another vector. The standard absorbing element is the absorbing
        element with norm 1.

        The absorbing element for circular convolution is the vector
        :math:`(1, 1, \dots, 1)^{\top} / \sqrt{d}`.

        Parameters
        ----------
        d : int
            Vector dimensionality.

        Returns
        -------
        (d,) ndarray
            Standard absorbing element.
        """
        return np.ones(d) / np.sqrt(d)

    def identity_element(self, d):
        r"""Return the identity element of dimensionality *d*.

        The identity does not change the vector it is bound to.

        The identity element for circular convolution is the vector
        :math:`(1, 0, \dots, 0)^{\top}`.

        Parameters
        ----------
        d : int
            Vector dimensionality.

        Returns
        -------
        (d,) ndarray
            Identity element.
        """
        data = np.zeros(d)
        data[0] = 1.0
        return data

    def zero_element(self, d):
        """Return the zero element of dimensionality *d*.

        The zero element produces itself when bound to a different vector.
        For circular convolution this is the zero vector.

        Parameters
        ----------
        d : int
            Vector dimensionality.

        Returns
        -------
        (d,) ndarray
            Zero element.
        """
        return np.zeros(d)



"""Generators to create vectors with specific properties."""

import numpy as np


def AxisAlignedVectors(d):
    """Generator for axis aligned vectors.

    Can yield at most *d* vectors.

    Note that while axis-aligned vectors can be useful for debugging,
    they will not work well with most binding methods for Semantic Pointers.

    Parameters
    ----------
    d : int
        Dimensionality of returned vectors.

    Examples
    --------
    >>> for p in nengo_spa.vector_generation.AxisAlignedVectors(4):
    ...    print(p)
    [1. 0. 0. 0.]
    [0. 1. 0. 0.]
    [0. 0. 1. 0.]
    [0. 0. 0. 1.]
    """
    for v in np.eye(d):
        yield v


class UnitLengthVectors:
    """Generator for uniformly distributed unit-length vectors.

    Parameters
    ----------
    d : int
        Dimensionality of returned vectors.
    rng : numpy.random.RandomState, optional
        The random number generator to use to create new vectors.
    """

    def __init__(self, d, rng=None):
        if rng is None:
            rng = np.random.RandomState()

        self.d = d
        self.rng = rng

    def __iter__(self):
        return self

    def __next__(self):
        v = self.rng.randn(self.d)
        v /= np.linalg.norm(v)
        return v

    def next(self):
        return self.__next__()


class UnitaryVectors:
    """Generator for unitary vectors (given some binding method).

    Parameters
    ----------
    d : int
        Dimensionality of returned vectors.
    algebra : AbstractAlgebra
        Algebra that defines what vectors are unitary.
    rng : numpy.random.RandomState, optional
        The random number generator to use to create new vectors.
    """

    def __init__(self, d, algebra, rng=None):
        if rng is None:
            rng = np.random.RandomState()

        self.d = d
        self.algebra = algebra
        self.rng = rng

    def __iter__(self):
        return self

    def __next__(self):
        return self.algebra.make_unitary(self.rng.randn(self.d))

    def next(self):
        return self.__next__()


class OrthonormalVectors:
    """Generator for random orthonormal vectors.

    Parameters
    ----------
    d : int
        Dimensionality of returned vectors.
    rng : numpy.random.RandomState, optional
        The random number generator to use to create new vectors.
    """

    def __init__(self, d, rng=None):
        if rng is None:
            rng = np.random.RandomState()

        self.d = d
        self.rng = rng
        self.vectors = []

    def __iter__(self):
        return self

    def __next__(self):
        v = self.rng.randn(self.d)
        i = len(self.vectors)
        if i >= self.d:
            raise StopIteration()
        elif i > 0:
            vectors = np.asarray(self.vectors)
            y = -np.dot(vectors[:, i:], v[i:])
            A = vectors[:i, :i]
            v[:i] = np.linalg.solve(A, y)
        v /= np.linalg.norm(v)
        self.vectors.append(v)
        return v

    def next(self):
        return self.__next__()


class ExpectedUnitLengthVectors:
    r"""Generator for vectors with expected unit-length.

    The vectors will be uniformly distributed with an expected norm
    of 1, but each specific pointer may have a length different than 1.
    Specifically each vector component will be normal distributed with mean 0
    and standard deviation :math:`1/\sqrt{d}`.

    Parameters
    ----------
    d : int
        Dimensionality of returned vectors.
    rng : numpy.random.RandomState, optional
        The random number generator to use to create new vectors.
    """

    def __init__(self, d, rng=None):
        if rng is None:
            rng = np.random.RandomState()

        self.d = d
        self.rng = rng

    def __iter__(self):
        return self

    def __next__(self):
        return self.rng.randn(self.d) / np.sqrt(self.d)

    def next(self):
        return self.__next__()



import pytest

from nengo_spa.ast.expr_tree import (
    BinaryOperator,
    Leaf,
    UnaryOperator,
)


def test_leaf_str():
    assert str(Leaf("foo")) == "foo"


def test_unary_operator_str():
    assert str(UnaryOperator("~", UnaryOperator("-", Leaf("leaf")))) == "~(-leaf)"
    assert (
        str(UnaryOperator("~", BinaryOperator("+", Leaf("a"), Leaf("b")))) == "~(a + b)"
    )
    assert (
        str(UnaryOperator("~", BinaryOperator("**", Leaf("a"), Leaf("b")))) == "~a ** b"
    )


def test_binary_operator_str():
    a, b, c = Leaf("a"), Leaf("b"), Leaf("c")
    assert str(BinaryOperator("+", a, BinaryOperator("*", b, c))) == "a + b * c"
    assert str(BinaryOperator("*", BinaryOperator("+", a, b), c)) == "(a + b) * c"
    assert str(BinaryOperator("-", BinaryOperator("+", a, b), c)) == "a + b - c"


@pytest.mark.parametrize("op", ["+", "-", "~"])
def test_unary_operator(op):
    leaf = Leaf("leaf")
    assert leaf
    assert str(eval(op + "leaf")) == op + "leaf"


@pytest.mark.parametrize("op", ["<<", ">>", "+", "-", "*", "@", "/", "//", "%", "**"])
def test_binary_operator(op):
    lhs, rhs = Leaf("lhs"), Leaf("rhs")
    assert lhs
    assert rhs
    assert str(eval("lhs" + op + "rhs")) == "lhs {} rhs".format(op)



import nengo
import numpy as np

from nengo_spa.algebras.base import AbstractAlgebra
from nengo_spa.networks.vtb import VTB


class VtbAlgebra(AbstractAlgebra):
    r"""Vector-derived Transformation Binding (VTB) algebra.

    VTB uses elementwise addition for superposition. The binding operation
    :math:`\mathcal{B}(x, y)` is defined as

    .. math::

       \mathcal{B}(x, y) := V_y x = \left[\begin{array}{ccc}
           V_y' &    0 &    0 \\
              0 & V_y' &    0 \\
              0 &    0 & V_y'
           \end{array}\right] x

    with

    .. math::

       V_y' = d^{\frac{1}{4}} \left[\begin{array}{cccc}
           y_1            & y_2            & \dots  & y_{d'}  \\
           y_{d' + 1}     & y_{d' + 2}     & \dots  & y_{2d'} \\
           \vdots         & \vdots         & \ddots & \vdots  \\
           y_{d - d' + 1} & y_{d - d' + 2} & \dots  & y_d
       \end{array}\right]

    and

    .. math:: d'^2 = d.

    The approximate inverse :math:`y^+` for :math:`y` is permuting the elements
    such that :math:`V_{y^+} = V_y`.

    Note that VTB requires the vector dimensionality to be square.

    The VTB binding operation is neither associative nor commutative.

    Publications with further information are forthcoming.
    """

    _instance = None

    def __new__(cls):
        if type(cls._instance) is not cls:
            cls._instance = super(VtbAlgebra, cls).__new__(cls)
        return cls._instance

    def is_valid_dimensionality(self, d):
        """Checks whether *d* is a valid vector dimensionality.

        For VTB all square numbers are valid dimensionalities.

        Parameters
        ----------
        d : int
            Dimensionality

        Returns
        -------
        bool
            *True*, if *d* is a valid vector dimensionality for the use with
            the algebra.
        """
        if d < 1:
            return False
        sub_d = np.sqrt(d)
        return sub_d * sub_d == d

    def _get_sub_d(self, d):
        sub_d = int(np.sqrt(d))
        if sub_d * sub_d != d:
            raise ValueError("Vector dimensionality must be a square number.")
        return sub_d

    def make_unitary(self, v):
        sub_d = self._get_sub_d(len(v))
        m = np.array(v.reshape((sub_d, sub_d)))
        for i in range(1, sub_d):
            y = -np.dot(m[:i, i:], m[i, i:])
            A = m[:i, :i]
            m[i, :i] = np.linalg.solve(A, y)
        m /= np.linalg.norm(m, axis=1)[:, None]
        m /= np.sqrt(sub_d)
        return m.flatten()

    def superpose(self, a, b):
        return a + b

    def bind(self, a, b):
        d = len(a)
        if len(b) != d:
            raise ValueError("Inputs must have same length.")
        m = self.get_binding_matrix(b)
        return np.dot(m, a)

    def invert(self, v):
        sub_d = self._get_sub_d(len(v))
        return v.reshape((sub_d, sub_d)).T.flatten()

    def get_binding_matrix(self, v, swap_inputs=False):
        sub_d = self._get_sub_d(len(v))
        m = np.sqrt(sub_d) * np.kron(np.eye(sub_d), v.reshape((sub_d, sub_d)))
        if swap_inputs:
            m = np.dot(self.get_swapping_matrix(len(v)), m)
        return m

    def get_swapping_matrix(self, d):
        """Get matrix to swap operands in bound state.

        Parameters
        ----------
        d : int
            Dimensionality of vector.

        Returns
        -------
        (d, d) ndarry
            Matrix to multiply with a vector to switch left and right operand
            in bound state.
        """
        return self.get_inversion_matrix(d)

    def get_inversion_matrix(self, d):
        sub_d = self._get_sub_d(d)
        return np.eye(d).reshape(d, sub_d, sub_d).T.reshape(d, d)

    def implement_superposition(self, n_neurons_per_d, d, n):
        node = nengo.Node(size_in=d)
        return node, n * (node,), node

    def implement_binding(self, n_neurons_per_d, d, unbind_left, unbind_right):
        net = VTB(n_neurons_per_d, d, unbind_left, unbind_right)
        return net, (net.input_left, net.input_right), net.output

    def absorbing_element(self, d):
        """VTB has no absorbing element except the zero vector."""
        raise NotImplementedError("VtbAlgebra does not have any absorbing elements.")

    def identity_element(self, d):
        sub_d = self._get_sub_d(d)
        return (np.eye(sub_d) / d ** 0.25).flatten()

    def zero_element(self, d):
        """Return the zero element of dimensionality *d*.

        The zero element produces itself when bound to a different vector.
        For VTB this is the zero vector.

        Parameters
        ----------
        d : int
            Vector dimensionality.

        Returns
        -------
        (d,) ndarray
            Zero element.
        """
        return np.zeros(d)



try:
    from pytest_nengo import (  # pylint: disable=unused-import
        pytest_configure,
        pytest_runtest_setup,
    )
except ImportError:
    import nengo
    import pytest

    @pytest.fixture(scope="session")
    def Simulator(request):
        return nengo.Simulator


from nengo_spa.algebras.hrr_algebra import HrrAlgebra
from nengo_spa.algebras.vtb_algebra import VtbAlgebra


class TestConfig:
    """Parameters affecting all NengoSPA testes.

    These are essentially global variables used by py.test to modify aspects
    of the NengoSPA tests. We collect them in this class to provide a mini
    namespace and to avoid using the ``global`` keyword.

    The values below are defaults. The functions in the remainder of this
    module modify these values accordingly.
    """

    algebras = [HrrAlgebra(), VtbAlgebra()]


def pytest_generate_tests(metafunc):
    if "algebra" in metafunc.fixturenames:
        metafunc.parametrize("algebra", [a for a in TestConfig.algebras])



import os

import pytest
import _pytest.capture

from nbformat import read as read_nb
from nengo.utils.stdlib import execfile

# Monkeypatch _pytest.capture.DontReadFromInput
#  If we don't do this, importing IPython will choke as it reads the current
#  sys.stdin to figure out the encoding it will use; pytest installs
#  DontReadFromInput as sys.stdin to capture output.
#  Running with -s option doesn't have this issue, but this monkeypatch
#  doesn't have any side effects, so it's fine.
_pytest.capture.DontReadFromInput.encoding = "utf-8"
_pytest.capture.DontReadFromInput.write = lambda: None
_pytest.capture.DontReadFromInput.flush = lambda: None

example_dir = "docs/examples"

too_slow = [
    "intro",
    "intro-coming-from-legacy-spa",
    "question",
    "question-control",
    "question-memory",
    "spa-parser",
    "spa-sequence",
    "spa-sequence-routed",
]

all_examples, slow_examples, fast_examples = [], [], []


def load_example(example):
    with open(example + ".ipynb", "r") as f:
        nb = read_nb(f, 4)
    return nb


for subdir, _, files in os.walk(example_dir):
    if (os.path.sep + ".") in subdir:
        continue
    files = [f for f in files if f.endswith(".ipynb")]
    examples = [os.path.join(subdir, os.path.splitext(f)[0]) for f in files]
    all_examples.extend(examples)
    slow_examples.extend(
        [e for e, f in zip(examples, files) if os.path.splitext(f)[0] in too_slow]
    )
    fast_examples.extend(
        [e for e, f in zip(examples, files) if os.path.splitext(f)[0] not in too_slow]
    )

# os.walk goes in arbitrary order, so sort after the fact to keep pytest happy
all_examples.sort()
slow_examples.sort()
fast_examples.sort()


def assert_noexceptions(nb_file, tmpdir):
    plt = pytest.importorskip("matplotlib.pyplot")
    pytest.importorskip("IPython", minversion="1.0")
    pytest.importorskip("jinja2")
    from nengo.utils.ipython import export_py

    nb = load_example(nb_file)
    pyfile = "%s.py" % tmpdir.join(os.path.basename(nb_file))
    export_py(nb, pyfile)
    execfile(pyfile, {})
    plt.close("all")


@pytest.mark.example
@pytest.mark.parametrize("nb_file", fast_examples)
def test_fast_noexceptions(nb_file, tmpdir):
    """Ensure that no cells raise an exception."""
    assert_noexceptions(nb_file, tmpdir)


@pytest.mark.slow
@pytest.mark.example
@pytest.mark.parametrize("nb_file", slow_examples)
def test_slow_noexceptions(nb_file, tmpdir):
    """Ensure that no cells raise an exception."""
    assert_noexceptions(nb_file, tmpdir)


def iter_cells(nb_file, cell_type="code"):
    nb = load_example(nb_file)

    if nb.nbformat <= 3:
        cells = []
        for ws in nb.worksheets:
            cells.extend(ws.cells)
    else:
        cells = nb.cells

    for cell in cells:
        if cell.cell_type == cell_type:
            yield cell


@pytest.mark.example
@pytest.mark.parametrize("nb_file", all_examples)
def test_no_signature(nb_file):
    nb = load_example(nb_file)
    assert "signature" not in nb.metadata, "Notebook has signature"


@pytest.mark.example
@pytest.mark.parametrize("nb_file", all_examples)
def test_no_outputs(nb_file):
    """Ensure that no cells have output."""
    pytest.importorskip("IPython", minversion="1.0")

    for cell in iter_cells(nb_file):
        assert cell.outputs == [], "Cell outputs not cleared"



# -*- coding: utf-8 -*-

from nengo.exceptions import NengoWarning, ValidationError
import nengo.solvers
import numpy as np
from numpy.testing import assert_equal
import pytest

from nengo_spa import SemanticPointer, Vocabulary
from nengo_spa.algebras import HrrAlgebra, VtbAlgebra
from nengo_spa.exceptions import SpaParseError
from nengo_spa.vector_generation import AxisAlignedVectors
from nengo_spa.vocabulary import (
    special_sps,
    VocabularyMap,
    VocabularyMapParam,
    VocabularyOrDimParam,
)


def test_add(rng):
    v = Vocabulary(3, pointer_gen=rng)
    v.add("A", [1, 2, 3])
    v.add("B", [4, 5, 6])
    v.add("C", [7, 8, 9])
    assert np.allclose(v.vectors, [[1, 2, 3], [4, 5, 6], [7, 8, 9]])


def test_add_raises_exception_for_algebra_mismatch():
    v = Vocabulary(4, algebra=HrrAlgebra())
    with pytest.raises(ValidationError, match="different vocabulary or algebra"):
        v.add("V", SemanticPointer(np.ones(4), algebra=VtbAlgebra()))
    v.add("V", SemanticPointer(np.ones(4), algebra=VtbAlgebra()).reinterpret(v))


def test_added_algebra_match(rng):
    v = Vocabulary(4, algebra=VtbAlgebra(), pointer_gen=rng)
    sp = v.create_pointer()
    assert sp.algebra is VtbAlgebra()
    v.add("V", sp)
    assert v["V"].vocab is v
    assert v["V"].algebra is VtbAlgebra()
    assert v["V"].name == "V"


def test_populate(rng):
    v = Vocabulary(64, pointer_gen=rng)

    v.populate("")
    v.populate(" \r\n\t")
    assert len(v) == 0

    v.populate("A")
    assert "A" in v

    v.populate("B; C")
    assert "B" in v
    assert "C" in v

    v.populate("D.unitary()")
    assert "D" in v
    np.testing.assert_almost_equal(np.linalg.norm(v["D"].v), 1.0)
    np.testing.assert_almost_equal(np.linalg.norm((v["D"] * v["D"]).v), 1.0)

    v.populate("E = A + 2 * B")
    assert np.allclose(v["E"].v, v.parse("A + 2 * B").v)
    assert np.linalg.norm(v["E"].v) > 2.0

    v.populate("F = (A + 2 * B).normalized()")
    assert np.allclose(v["F"].v, v.parse("A + 2 * B").normalized().v)
    np.testing.assert_almost_equal(np.linalg.norm(v["F"].v), 1.0)

    v.populate("G = A; H")
    assert np.allclose(v["G"].v, v["A"].v)
    assert "H" in v

    # Assigning non-existing pointer
    with pytest.raises(NameError):
        v.populate("I = J")

    # Redefining
    with pytest.raises(ValidationError):
        v.populate("H = A")

    # Calling non existing function
    with pytest.raises(AttributeError):
        v.populate("I = H.invalid()")

    # invalid names: lowercase, unicode
    with pytest.raises(SpaParseError):
        v.populate("x = A")
    with pytest.raises(SpaParseError):
        v.populate(u"Aα = A")


def test_pointer_gen():
    v = Vocabulary(32, pointer_gen=AxisAlignedVectors(32))
    v.populate("A; B; C")
    assert np.all(v.vectors == np.eye(32)[:3])


@pytest.mark.parametrize("pointer_gen", ("string", 123))
def test_invalid_pointer_gen(pointer_gen):
    with pytest.raises(ValidationError):
        Vocabulary(32, pointer_gen=pointer_gen)


@pytest.mark.parametrize(
    "name", ("None", "True", "False", "Zero", "AbsorbingElement", "Identity")
)
def test_reserved_names(name):
    v = Vocabulary(16)
    with pytest.raises(SpaParseError):
        v.populate(name)


@pytest.mark.parametrize("name,sp", sorted(special_sps.items()))
def test_special_sps(name, sp, rng):
    v = Vocabulary(16, pointer_gen=rng)
    assert name in v
    assert np.allclose(v[name].v, sp(16).v)
    assert np.allclose(v.parse(name).v, sp(16).v)


def test_populate_with_transform_on_first_vector(rng):
    v = Vocabulary(64, pointer_gen=rng)

    v.populate("A.unitary()")
    assert "A" in v
    assert np.allclose(v["A"].v, v["A"].unitary().v)


def test_populate_with_transform_on_nonstrict_vocab(rng):
    v = Vocabulary(64, pointer_gen=rng, strict=False)

    v.populate("A.unitary()")
    assert "A" in v
    assert np.allclose(v["A"].v, v["A"].unitary().v)


def test_parse(rng):
    v = Vocabulary(64, pointer_gen=rng)
    v.populate("A; B; C")
    A = v.parse("A")
    B = v.parse("B")
    C = v.parse("C")
    assert np.allclose((A * B).v, v.parse("A * B").v)
    assert np.allclose((A * ~B).v, v.parse("A * ~B").v)
    assert np.allclose((A + B).v, v.parse("A + B").v)
    assert np.allclose((A - (B * C) * 3 + ~C).v, v.parse("A-(B*C)*3+~C").v)

    assert np.allclose(v.parse("0").v, np.zeros(64))
    assert np.allclose(v.parse("1").v, np.eye(64)[0])
    assert np.allclose(v.parse("1.7").v, np.eye(64)[0] * 1.7)

    with pytest.raises(SyntaxError):
        v.parse("A((")
    with pytest.raises(SpaParseError):
        v.parse('"hello"')
    with pytest.raises(SpaParseError):
        v.parse('"hello"')


def test_parse_n(rng):
    v = Vocabulary(64, pointer_gen=rng)
    v.populate("A; B; C")
    A = v.parse("A")
    B = v.parse("B")

    parsed = v.parse_n("A", "A*B", "A+B", "3")
    assert np.allclose(parsed[0].v, A.v)
    assert np.allclose(parsed[1].v, (A * B).v)
    assert np.allclose(parsed[2].v, (A + B).v)
    # FIXME should give an exception?
    assert np.allclose(parsed[3].v, 3 * v["Identity"].v)


def test_invalid_dimensions():
    with pytest.raises(ValidationError):
        Vocabulary(1.5)
    with pytest.raises(ValidationError):
        Vocabulary(0)
    with pytest.raises(ValidationError):
        Vocabulary(-1)


def test_capital(rng):
    v = Vocabulary(16, pointer_gen=rng)
    with pytest.raises(SpaParseError):
        v.parse("a")
    with pytest.raises(SpaParseError):
        v.parse("A+B+C+a")


@pytest.mark.parametrize("solver", [None, nengo.solvers.Lstsq()])
def test_transform(rng, solver):
    v1 = Vocabulary(32, strict=False, pointer_gen=rng)
    v2 = Vocabulary(64, strict=False, pointer_gen=rng)
    v1.populate("A; B; C")
    v2.populate("A; B; C")
    A = v1["A"]
    B = v1["B"]
    C = v1["C"]

    # Test transform from v1 to v2 (full vocbulary)
    # Expected: np.dot(t, A.v) ~= v2.parse('A')
    # Expected: np.dot(t, B.v) ~= v2.parse('B')
    # Expected: np.dot(t, C.v) ~= v2.parse('C')
    t = v1.transform_to(v2, solver=solver)

    assert v2.parse("A").compare(np.dot(t, A.v)) > 0.85
    assert v2.parse("C+B").compare(np.dot(t, C.v + B.v)) > 0.85

    # Test transform from v1 to v2 (only 'A' and 'B')
    t = v1.transform_to(v2, keys=["A", "B"], solver=solver)

    assert v2.parse("A").compare(np.dot(t, A.v)) > 0.85
    assert v2.parse("B").compare(np.dot(t, C.v + B.v)) > 0.85

    # Test warns on missing keys
    v1.populate("D")
    D = v1["D"]
    with pytest.warns(NengoWarning):
        v1.transform_to(v2, solver=solver)

    # Test populating missing keys
    t = v1.transform_to(v2, populate=True, solver=solver)
    assert v2.parse("D").compare(np.dot(t, D.v)) > 0.85

    # Test ignores missing keys in source vocab
    v2.populate("E")
    v1.transform_to(v2, populate=True, solver=solver)
    assert "E" not in v1


def test_create_pointer_warning(rng):
    v = Vocabulary(2, pointer_gen=rng)

    # five pointers shouldn't fit
    with pytest.warns(UserWarning):
        v.populate("A; B; C; D; E")


def test_readonly(rng):
    v1 = Vocabulary(32, pointer_gen=rng)
    v1.populate("A;B;C")

    v1.readonly = True

    with pytest.raises(ValueError):
        v1.parse("D")


def test_subset(rng, algebra):
    v1 = Vocabulary(32, pointer_gen=rng, algebra=algebra)
    v1.populate("A; B; C; D; E; F; G")

    # Test creating a vocabulary subset
    v2 = v1.create_subset(["A", "C", "E"])
    assert list(v2.keys()) == ["A", "C", "E"]
    assert_equal(v2["A"].v, v1["A"].v)
    assert_equal(v2["C"].v, v1["C"].v)
    assert_equal(v2["E"].v, v1["E"].v)

    assert v1.algebra is v2.algebra


def test_vocabulary_tracking(rng):
    v = Vocabulary(32, pointer_gen=rng)
    v.populate("A")

    assert v["A"].vocab is v
    assert v.parse("2 * A").vocab is v


def test_vocabulary_set(rng):
    v8 = Vocabulary(8)
    v16 = Vocabulary(16)
    v32 = Vocabulary(32)
    vs = VocabularyMap([v8, v16], rng=rng)

    # Behaviour common to set and dict
    assert len(vs) == 2
    assert 8 in vs
    assert 16 in vs
    assert 32 not in vs

    assert v8 in vs
    assert v16 in vs
    assert v32 not in vs
    assert Vocabulary(8) not in vs

    # dict behaviour
    assert vs[8] is v8
    assert vs[16] is v16

    del vs[8]
    assert 8 not in vs

    # set behaviour
    vs.add(v32)
    assert vs[32] is v32
    with pytest.warns(UserWarning):
        vs.add(v32)

    vs.discard(32)
    assert 32 not in vs
    vs.discard(v16)
    assert 16 not in vs

    # creating new vocabs if non existent
    vs.add(v8)
    assert vs.get_or_create(8) is v8
    new = vs.get_or_create(16)
    assert vs[16] is new
    assert new.dimensions == 16
    assert new.pointer_gen.rng is rng


def test_vocabulary_map_param():
    class Test:
        vocab_map = VocabularyMapParam("vocab_map", readonly=False)

    obj = Test()
    vm = VocabularyMap()
    v16 = Vocabulary(16)
    v32 = Vocabulary(32)

    obj.vocab_map = vm
    assert obj.vocab_map is vm

    obj.vocab_map = [v16, v32]
    assert obj.vocab_map[16] is v16
    assert obj.vocab_map[32] is v32

    with pytest.raises(ValidationError):
        obj.vocab_map = "incompatible"


def test_vocabulary_or_dim_param():
    v16 = Vocabulary(16)
    v32 = Vocabulary(32)

    class Test:
        vocabs = VocabularyMap([v16])
        vocab = VocabularyOrDimParam("vocab", readonly=False)

    obj = Test()

    obj.vocab = v32
    assert obj.vocab is v32

    obj.vocab = 16
    assert obj.vocab is v16

    with pytest.raises(ValidationError):
        obj.vocab = "incompatible"

    with pytest.raises(ValidationError):
        obj.vocab = 0


def test_pointer_names():
    v = Vocabulary(16)
    v.populate("A; B")

    assert v["A"].name == "A"
    assert v.parse("A*B").name == "(A)*(B)"



import numpy as np
import pytest

from nengo_spa.algebras import HrrAlgebra
from nengo_spa.semantic_pointer import SemanticPointer
from nengo_spa.vector_generation import (
    AxisAlignedVectors,
    ExpectedUnitLengthVectors,
    OrthonormalVectors,
    UnitLengthVectors,
    UnitaryVectors,
)


def test_axis_aligned_pointers():
    for i, p in enumerate(AxisAlignedVectors(5)):
        assert np.all(p == np.eye(5)[i])


def test_unit_length_pointers(rng):
    g = UnitLengthVectors(64, rng=rng)
    for i in range(10):
        assert np.allclose(np.linalg.norm(next(g)), 1.0)


def test_unitary_pointers(rng):
    algebra = HrrAlgebra()
    g = UnitaryVectors(64, algebra, rng=rng)
    a = SemanticPointer(next(g), algebra=algebra)
    b = SemanticPointer(next(g), algebra=algebra)
    c = SemanticPointer(next(g), algebra=algebra)
    assert np.allclose(a.compare(c), (a * b).compare(c * b))


def test_orthonormal_pointers(rng):
    g = OrthonormalVectors(32, rng=rng)
    vectors = np.array(list(g))
    assert len(vectors) == 32
    assert np.allclose(np.dot(vectors.T, vectors), np.eye(32))


def test_expected_unit_length_vectors(rng):
    g = ExpectedUnitLengthVectors(64, rng=rng)
    assert np.abs(np.mean([np.linalg.norm(next(g)) for i in range(50)]) - 1.0) < 0.1


@pytest.mark.parametrize(
    "vg",
    (
        AxisAlignedVectors,
        ExpectedUnitLengthVectors,
        OrthonormalVectors,
        UnitLengthVectors,
        lambda d: UnitaryVectors(d, algebra=HrrAlgebra()),
    ),
)
def test_instantiation_without_rng(vg):
    d = 64
    assert len(next(vg(d))) == d


@pytest.mark.parametrize(
    "vg",
    (
        AxisAlignedVectors,
        ExpectedUnitLengthVectors,
        OrthonormalVectors,
        UnitLengthVectors,
        lambda d: UnitaryVectors(d, algebra=HrrAlgebra()),
    ),
)
def test_iter(vg):
    d = 64
    x = vg(d)
    assert iter(x) is x



import re

import numpy as np
from numpy.testing import assert_almost_equal

from nengo_spa.examine import pairs, similarity, text
from nengo_spa.vocabulary import Vocabulary


def test_similarity(rng):
    v = Vocabulary(64, max_similarity=0.1, pointer_gen=rng)
    v.populate(
        """
        A;
        B;
        C = .7 * A + .3 * B;
        D = C.normalized()
    """
    )

    s = similarity(v["A"].v, v)
    assert_almost_equal(s[0], 1.0)
    assert s[1] < 0.1
    assert_almost_equal(s[2], 0.7 + 0.3 * np.dot(v["A"].v, v["B"].v))

    sn = similarity(v["A"].v, v, normalize=True)
    assert_almost_equal(s[3], sn[2])

    data = np.array([v["A"].v, v["B"].v])
    s = similarity(data, v)
    assert_almost_equal(s[0, 0], 1.0)
    assert_almost_equal(s[1, 1], 1.0)
    assert s[0, 1] < 0.1
    assert s[1, 0] < 0.1

    s = similarity(data, v.parse_n("A", "B"))
    assert_almost_equal(s[0, 0], 1.0)
    assert_almost_equal(s[1, 1], 1.0)
    assert s[0, 1] < 0.1
    assert s[1, 0] < 0.1


def test_pairs():
    v = Vocabulary(64)
    v.populate("A; B; C")
    actual = pairs(v)
    expected = {"A*B", "A*C", "B*C"}
    assert actual == expected


def test_text(rng):
    v = Vocabulary(64, pointer_gen=rng)
    v.populate("A; B; C; D; E; F")
    x = v.parse("A+B+C")
    y = v.parse("-D-E-F")
    ptr = r"-?[01]\.[0-9]{2}[A-F]"
    assert re.match(";".join([ptr] * 3), text(x, v))
    assert re.match(";".join([ptr] * 2), text(x, v, maximum_count=2))
    assert re.match(ptr, text(x, v, maximum_count=1))
    assert len(text(x, v, maximum_count=10).split(";")) <= 10
    assert re.match(";".join([ptr] * 4), text(x, v, minimum_count=4))
    assert re.match(
        ";".join([ptr.replace("F", "C")] * 3),
        text(x, v, minimum_count=4, terms=["A", "B", "C"]),
    )

    assert re.match(ptr, text(y, v, threshold=0.6))
    assert text(y, v, minimum_count=None, threshold=0.6) == ""

    assert text(x, v, join=",") == text(x, v).replace(";", ",")
    assert re.match(";".join([ptr] * 2), text(x, v, normalize=True))

    assert text([0] * 64, v) == "0.00F"
    assert text(v["D"].v, v) == "1.00D"



"""Selection networks that pick one or more options among multiple choices."""
import nengo
import numpy as np


class IA(nengo.Network):
    """Independent accumulator (IA) winner-take-all (WTA) network.

    This is a two-layered network. The first layer consists of independent
    accumulators (integrators), whereas the second layer does a thresholding.
    Once the threshold is exceeded a feedback connection will stabilize the
    current choice and inhibit all other choices. To switch the selection,
    it is necessary to provide a transient input to *input_reset* to reset
    the accumulator states.

    This network is suited especially for accumulating evidence under noisy
    conditions and keep a stable choice selection until the processing of the
    choice has been finished.

    Further details are to be found in [gosmann2017]_.

    Parameters
    ----------
    n_neurons : int
        Number of neurons for each choice.
    n_ensembles : int
        Number of choices.
    accum_threshold : float, optional
        Accumulation threshold that needs to be reached to produce an output.
    accum_neuron_ratio: float, optional
        Portion of *n_neurons* that will be used for a layer 1 accumulator
        ensemble. The remaining neurons will be used for a layer 2 thresholding
        ensemble.
    accum_timescale : float, optional
        Evidence accumulation timescale.
    feedback_timescale : float, optional
        Timescale for the feedback connection from the thresholding layer to
        the accumulation layer.
    accum_synapse : Synapse or float, optional
        The synapse for connections to the accumulator ensembles.
    ff_synapse : Synapse or float, optional
        Synapse for feed-forward connections.
    intercept_width : float, optional
        The `nengo.presets.ThresholdingEnsembles` *intercept_width* parameter.
    radius : float, optional
        The representational radius of the ensembles.
    **kwargs : dict
        Keyword arguments passed on to `nengo.Network`.

    Attributes
    ----------
    input : nengo.Node
        The inputs to the network.
    input_reset : nengo.Node
        Input to reset the accumulators.
    output : nengo.Node
        The outputs of the network.
    accumulators : nengo.Thresholding
        The layer 1 accumulators.
    thresholding : nengo.Thresholding
        The layer 2 thresholding ensembles.

    References
    ----------
    .. [gosmann2017] Jan Gosmann, Aaron R. Voelker, and Chris Eliasmith. "A
       spiking independent accumulator model for winner-take-all computation."
       In Proceedings of the 39th Annual Conference of the Cognitive Science
       Society. London, UK, 2017. Cognitive Science Society.
    """

    def __init__(
        self,
        n_neurons,
        n_ensembles,
        accum_threshold=0.8,
        accum_neuron_ratio=0.7,
        accum_timescale=0.2,
        feedback_timescale=0.005,
        accum_synapse=0.1,
        ff_synapse=0.005,
        intercept_width=0.15,
        radius=1.0,
        **kwargs
    ):
        super().__init__(**kwargs)

        n_accum_neurons = int(accum_neuron_ratio * n_neurons)
        n_thresholding_neurons = n_neurons - n_accum_neurons

        bar_beta = 1.0 + radius * feedback_timescale / accum_timescale
        feedback_tr = (
            np.eye(n_ensembles)
            - bar_beta * (1.0 - np.eye(n_ensembles)) / feedback_timescale
        )

        with self:
            self.accumulators = Thresholding(
                n_accum_neurons,
                n_ensembles,
                threshold=0.0,
                intercept_width=intercept_width,
                radius=radius,
            )
            self.thresholding = Thresholding(
                n_thresholding_neurons,
                n_ensembles,
                threshold=accum_threshold,
                intercept_width=intercept_width,
                radius=radius,
                function=lambda x: x > accum_threshold,
            )

            nengo.Connection(
                self.accumulators.output, self.accumulators.input, synapse=accum_synapse
            )
            nengo.Connection(
                self.accumulators.output, self.thresholding.input, synapse=ff_synapse
            )
            nengo.Connection(
                self.thresholding.output,
                self.accumulators.input,
                synapse=accum_synapse,
                transform=accum_synapse * feedback_tr,
            )

            self.input_reset = nengo.Node(size_in=1)
            nengo.Connection(
                self.input_reset,
                self.accumulators.input,
                synapse=None,
                transform=-radius * np.ones((n_ensembles, 1)) / accum_synapse,
            )

            self.input = nengo.Node(size_in=n_ensembles)
            nengo.Connection(
                self.input,
                self.accumulators.input,
                synapse=None,
                transform=1.0 / accum_timescale,
            )
            self.output = self.thresholding.output


class Thresholding(nengo.Network):
    """Array of thresholding ensembles.

    All inputs below the threshold will produce an output of 0, whereas inputs
    above the threshold produce an output of equal value.

    Parameters
    ----------
    n_neurons : int
        Number of neurons for each ensemble.
    n_ensembles : int
        Number of ensembles.
    threshold : float
        The thresholding value.
    intercept_width : float, optional
        The `nengo.presets.ThresholdingEnsembles` *intercept_width* parameter.
    function : function, optional
        Function to apply to the thresholded values.
    radius : float, optional
        The representational radius of the ensembles.
    **kwargs : dict
        Keyword arguments passed on to `nengo.Network`.

    Attributes
    ----------
    input : nengo.Node
        The inputs to the network.
    output : nengo.Node
        The outputs of the network.
    thresholded : nengo.Node
        The raw thresholded value (before applying *function* or correcting for
        the shift produced by the thresholding).
    """

    def __init__(
        self,
        n_neurons,
        n_ensembles,
        threshold,
        intercept_width=0.15,
        function=None,
        radius=1.0,
        **kwargs
    ):
        super().__init__(**kwargs)

        with self:
            with nengo.presets.ThresholdingEnsembles(
                0.0, intercept_width, radius=radius
            ):
                self.thresholding = nengo.networks.EnsembleArray(n_neurons, n_ensembles)

            self.bias = nengo.Node(1.0)
            nengo.Connection(
                self.bias,
                self.thresholding.input,
                transform=-threshold * np.ones((n_ensembles, 1)),
            )

            self.input = self.thresholding.input
            self.thresholded = self.thresholding.output

            if function is None:
                function = lambda x: x
            function = lambda x, function=function: function(x + threshold)
            self.output = self.thresholding.add_output("function", function)


class WTA(Thresholding):
    """Winner-take-all (WTA) network with lateral inhibition.

    Parameters
    ----------
    n_neurons : int
        Number of neurons for each ensemble.
    n_ensembles : int
        Number of ensembles.
    inhibit_scale : float, optional
        Scaling of the lateral inhibition.
    inhibit_synapse : Synapse or float, optional
        Synapse on the recurrent connection for lateral inhibition.
    **kwargs : dict
        Keyword arguments passed on to `Thresholding`.

    Attributes
    ----------
    input : nengo.Node
        The inputs to the network.
    output : nengo.Node
        The outputs of the network.
    thresholded : nengo.Node
        The raw thresholded value (before applying *function* or correcting for
        the shift produced by the thresholding).
    """

    def __init__(
        self, n_neurons, n_ensembles, inhibit_scale=1.0, inhibit_synapse=0.005, **kwargs
    ):
        super().__init__(n_neurons, n_ensembles, **kwargs)

        with self:
            nengo.Connection(
                self.thresholded,
                self.input,
                transform=inhibit_scale * (np.eye(n_ensembles) - 1.0),
                synapse=inhibit_synapse,
            )



import sys

import nengo
import numpy as np
from numpy.testing import assert_allclose
import pytest

import nengo_spa as spa
from nengo_spa.ast.symbolic import PointerSymbol
from nengo_spa.exceptions import SpaTypeError
from nengo_spa.semantic_pointer import SemanticPointer
from nengo_spa.testing import assert_sp_close


def test_scalar_addition(Simulator, seed):
    with spa.Network(seed=seed) as model:
        a = spa.Scalar()
        b = spa.Scalar()

        0.3 >> a
        0.2 >> b
        n1 = (a + b).construct()
        n2 = nengo.Node(size_in=1)
        (a + b).connect_to(n2)

        p1 = nengo.Probe(n1, synapse=0.03)
        p2 = nengo.Probe(n2, synapse=0.03)

    with Simulator(model) as sim:
        sim.run(0.3)

    assert np.all(np.abs(sim.data[p1][sim.trange() > 0.2] - 0.5) < 0.1)
    assert np.all(np.abs(sim.data[p2][sim.trange() > 0.2] - 0.5) < 0.1)


@pytest.mark.parametrize("op", ["-", "~"])
@pytest.mark.parametrize("suffix", ["", ".output"])
def test_unary_operation_on_module(Simulator, algebra, op, suffix, rng):
    vocab = spa.Vocabulary(16, pointer_gen=rng, algebra=algebra)
    vocab.populate("A")

    with spa.Network() as model:
        stimulus = spa.Transcode("A", output_vocab=vocab)  # noqa: F841
        x = eval(op + "stimulus" + suffix)
        p = nengo.Probe(x.construct(), synapse=0.03)

    with Simulator(model) as sim:
        sim.run(0.3)

    assert_sp_close(sim.trange(), sim.data[p], vocab.parse(op + "A"), skip=0.2)


@pytest.mark.parametrize("op", ["+", "-", "*"])
@pytest.mark.parametrize("suffix", ["", ".output"])
def test_binary_operation_on_modules(Simulator, algebra, op, suffix, seed, rng):
    vocab = spa.Vocabulary(16, pointer_gen=rng, algebra=algebra)
    vocab.populate("A; B")

    with spa.Network(seed=seed) as model:
        a = spa.Transcode("A", output_vocab=vocab)  # noqa: F841
        b = spa.Transcode("B", output_vocab=vocab)  # noqa: F841
        x = eval("a" + suffix + op + "b" + suffix)
        p = nengo.Probe(x.construct(), synapse=0.03)

    with Simulator(model) as sim:
        sim.run(0.3)

    assert_sp_close(
        sim.trange(), sim.data[p], vocab.parse("A" + op + "B"), skip=0.2, atol=0.3
    )


@pytest.mark.parametrize("suffix", ["", ".output"])
def test_division_with_fixed(Simulator, suffix, seed, rng):
    vocab = spa.Vocabulary(16, pointer_gen=rng)
    vocab.populate("A")

    with spa.Network(seed=seed) as model:
        a = spa.Transcode("A", output_vocab=vocab)  # noqa: F841
        x = eval("a" + suffix + "/ 2")
        p = nengo.Probe(x.construct(), synapse=0.03)

    with Simulator(model) as sim:
        sim.run(0.3)

    assert_sp_close(
        sim.trange(), sim.data[p], vocab.parse("0.5 * A"), skip=0.2, atol=0.3
    )


@pytest.mark.parametrize("op", ["+", "-", "*"])
@pytest.mark.parametrize("order", ["AB", "BA"])
def test_binary_operation_on_modules_with_pointer_symbol(
    Simulator, algebra, op, order, seed, rng
):
    vocab = spa.Vocabulary(16, pointer_gen=rng, algebra=algebra)
    vocab.populate("A; B")

    with spa.Network(seed=seed) as model:
        a = spa.Transcode("A", output_vocab=vocab)  # noqa: F841
        if order == "AB":
            x = eval("a" + op + 'PointerSymbol("B")')
        elif order == "BA":
            x = eval('PointerSymbol("B")' + op + "a")
        else:
            raise ValueError("Invalid order argument.")
        p = nengo.Probe(x.construct(), synapse=0.03)

    with Simulator(model) as sim:
        sim.run(0.3)

    assert_sp_close(
        sim.trange(), sim.data[p], vocab.parse(order[0] + op + order[1]), skip=0.2
    )


@pytest.mark.parametrize("op", ["+", "-", "*"])
@pytest.mark.parametrize("order", ["AB", "BA"])
def test_binary_operation_on_modules_with_fixed_pointer(
    Simulator, algebra, op, order, seed, rng
):
    vocab = spa.Vocabulary(16, pointer_gen=rng, algebra=algebra)
    vocab.populate("A; B")
    b = SemanticPointer(vocab["B"].v)  # noqa: F841

    with spa.Network(seed=seed) as model:
        a = spa.Transcode("A", output_vocab=vocab)  # noqa: F841
        if order == "AB":
            x = eval("a" + op + "b")
        elif order == "BA":
            x = eval("b" + op + "a")
        else:
            raise ValueError("Invalid order argument.")
        p = nengo.Probe(x.construct(), synapse=0.03)

    with Simulator(model) as sim:
        sim.run(0.3)

    assert_sp_close(
        sim.trange(),
        sim.data[p],
        vocab.parse(order[0] + op + order[1]),
        skip=0.2,
        atol=0.3,
    )


def test_complex_rule(Simulator, algebra, rng):
    vocab = spa.Vocabulary(16, pointer_gen=rng, algebra=algebra)
    vocab.populate("A; B; C; D")

    with spa.Network() as model:
        a = spa.Transcode("A", output_vocab=vocab)
        b = spa.Transcode("B", output_vocab=vocab)

        x = (0.5 * PointerSymbol("C") * a + 0.5 * PointerSymbol("D")) * (
            0.5 * b + a * 0.5
        )
        p = nengo.Probe(x.construct(), synapse=0.3)

    with Simulator(model) as sim:
        sim.run(0.3)

    assert_sp_close(
        sim.trange(),
        sim.data[p],
        vocab.parse("(0.5 * C * A + 0.5 * D) * (0.5 * B + 0.5 * A)"),
        skip=0.2,
        normalized=True,
    )


def test_transformed(Simulator, algebra, seed, rng):
    vocab = spa.Vocabulary(16, pointer_gen=rng, algebra=algebra)
    vocab.populate("A; B")

    with spa.Network(seed=seed) as model:
        a = spa.Transcode("A", output_vocab=vocab)
        x = PointerSymbol("B") * a
        p = nengo.Probe(x.construct(), synapse=0.3)

    with Simulator(model) as sim:
        sim.run(0.3)

    assert_sp_close(
        sim.trange(), sim.data[p], vocab.parse("B*A"), skip=0.2, normalized=True
    )


def test_transformed_and_pointer_symbol(Simulator, algebra, seed, rng):
    vocab = spa.Vocabulary(16, pointer_gen=rng, algebra=algebra)
    vocab.populate("A; B")

    with spa.Network(seed=seed) as model:
        a = spa.Transcode("A", output_vocab=vocab)
        x = (a * PointerSymbol("B")) * PointerSymbol("~B")
        p = nengo.Probe(x.construct(), synapse=0.3)

    with Simulator(model) as sim:
        sim.run(0.3)

    assert_sp_close(
        sim.trange(), sim.data[p], vocab.parse("A * B * ~B"), skip=0.2, normalized=True
    )


def test_transformed_and_network(Simulator, algebra, seed, rng):
    vocab = spa.Vocabulary(16, pointer_gen=rng, algebra=algebra)
    vocab.populate("A; B.unitary()")

    with spa.Network(seed=seed) as model:
        a = spa.Transcode("A", output_vocab=vocab)
        b = spa.Transcode("B", output_vocab=vocab)
        x = (a * PointerSymbol("~B")) * b
        p = nengo.Probe(x.construct(), synapse=0.3)

    with Simulator(model) as sim:
        sim.run(0.3)

    assert_sp_close(
        sim.trange(), sim.data[p], vocab.parse("A * ~B * B"), skip=0.2, normalized=True
    )


def test_transformed_and_transformed(Simulator, algebra, seed, rng):
    vocab = spa.Vocabulary(16, pointer_gen=rng, algebra=algebra)
    vocab.populate("A; B.unitary(); C")

    with spa.Network(seed=seed) as model:
        a = spa.Transcode("A", output_vocab=vocab)
        c = spa.Transcode("C", output_vocab=vocab)
        x = (PointerSymbol("B") * a) * (PointerSymbol("~B") * c)
        p = nengo.Probe(x.construct(), synapse=0.3)

    with Simulator(model) as sim:
        sim.run(0.3)

    assert_sp_close(
        sim.trange(),
        sim.data[p],
        vocab.parse("(B * A) * (~B * C)"),
        skip=0.2,
        normalized=True,
        atol=0.3,
    )


def test_pointer_symbol_with_dynamic_scalar():
    with spa.Network():
        scalar = spa.Scalar()
        with pytest.raises(SpaTypeError):
            PointerSymbol("A") * scalar


def test_dot(Simulator, seed, rng):
    vocab = spa.Vocabulary(16, pointer_gen=rng)
    vocab.populate("A; B")

    with spa.Network(seed=seed) as model:
        a = spa.Transcode("A", output_vocab=vocab)
        b = spa.Transcode(lambda t: "A" if t <= 0.5 else "B", output_vocab=vocab)
        x = spa.dot(a, b)
        p = nengo.Probe(x.construct(), synapse=0.03)

    with Simulator(model) as sim:
        sim.run(1.0)

    t = sim.trange()
    assert_allclose(sim.data[p][(0.3 < t) & (t <= 0.5)], 1.0, atol=0.2)
    assert np.all(sim.data[p][0.8 < t] < 0.2)


@pytest.mark.parametrize("a", ("PointerSymbol('A')", "vocab['A']"))
def test_dot_with_fixed(Simulator, seed, rng, a):
    vocab = spa.Vocabulary(16, pointer_gen=rng)
    vocab.populate("A; B")

    with spa.Network(seed=seed) as model:
        a = eval(a)
        b = spa.Transcode(lambda t: "A" if t <= 0.5 else "B", output_vocab=vocab)
        network_count = len(model.all_networks)
        x = spa.dot(a, b)
        # transform should suffice, no new networks should be created
        assert len(model.all_networks) == network_count
        p = nengo.Probe(x.construct(), synapse=0.03)

    with Simulator(model) as sim:
        sim.run(1.0)

    t = sim.trange()
    assert_allclose(sim.data[p][(0.3 < t) & (t <= 0.5)], 1.0, atol=0.2)
    assert np.all(sim.data[p][0.8 < t] < 0.2)


@pytest.mark.skipif(sys.version_info < (3, 5), reason="requires Python 3.5")
def test_dot_matmul(Simulator, seed, rng):
    vocab = spa.Vocabulary(16, pointer_gen=rng)
    vocab.populate("A; B")

    with spa.Network(seed=seed) as model:
        a = spa.Transcode("A", output_vocab=vocab)  # noqa: F841
        b = spa.Transcode(  # noqa: F841
            lambda t: "A" if t <= 0.5 else "B", output_vocab=vocab
        )
        x = eval("a @ b")
        p = nengo.Probe(x.construct(), synapse=0.03)

    with Simulator(model) as sim:
        sim.run(1.0)

    t = sim.trange()
    assert_allclose(sim.data[p][(0.3 < t) & (t <= 0.5)], 1.0, atol=0.2)
    assert np.all(sim.data[p][0.8 < t] < 0.2)


@pytest.mark.skipif(sys.version_info < (3, 5), reason="requires Python 3.5")
def test_dot_with_fixed_matmul(Simulator, seed, rng):
    vocab = spa.Vocabulary(16, pointer_gen=rng)
    vocab.populate("A; B")

    with spa.Network(seed=seed) as model:
        a = PointerSymbol("A")  # noqa: F841
        b = spa.Transcode(  # noqa: F841
            lambda t: "A" if t <= 0.5 else "B", output_vocab=vocab
        )
        x = eval("a @ b")
        p = nengo.Probe(x.construct(), synapse=0.03)

    with Simulator(model) as sim:
        sim.run(1.0)

    t = sim.trange()
    assert_allclose(sim.data[p][(0.3 < t) & (t <= 0.5)], 1.0, atol=0.2)
    assert np.all(sim.data[p][0.8 < t] < 0.2)


def test_dynamic_translate(Simulator, seed, rng):
    v1 = spa.Vocabulary(64, pointer_gen=rng)
    v1.populate("A; B")
    v2 = spa.Vocabulary(64, pointer_gen=rng)
    v2.populate("A; B")

    with spa.Network(seed=seed) as model:
        source = spa.Transcode("A", output_vocab=v1)
        x = spa.translate(source, v2)
        p = nengo.Probe(x.construct(), synapse=0.03)

    with Simulator(model) as sim:
        sim.run(0.5)

    assert_sp_close(sim.trange(), sim.data[p], v2["A"], skip=0.3, atol=0.2)



"""Representation of simple expression trees."""

from collections import namedtuple


# Reference precedence table:
# https://docs.python.org/3/reference/expressions.html
precedence = {
    op: p
    for p, operators in enumerate(
        (
            {":="},
            {"lambda"},
            {"if", "else"},
            {"or"},
            {"and"},
            {"not x"},
            {"in", "not in", "is", "is not", "<", "<=", ">", ">=", "!=", "=="},
            {"|"},
            {"^"},
            {"&"},
            {"<<", ">>"},
            {"+", "-"},
            {"*", "@", "/", "//", "%"},
            {"+x", "-x", "~x"},
            {"**"},
            {"await x"},
            {"x[index]", "x[index:index]", "x(arguments...)", "x.attribute"},
            {
                "(expressions...)",
                "[expressions...]",
                "{key: value...}",
                "{expressions...}",
            },
        )
    )
    for op in operators
}


class Node(namedtuple("Node", ["value", "precedence", "children"])):
    def __invert__(self):
        return UnaryOperator("~", self)

    def __neg__(self):
        return UnaryOperator("-", self)

    def __pos__(self):
        return UnaryOperator("+", self)

    def __add__(self, other):
        return BinaryOperator("+", self, other)

    def __radd__(self, other):
        return BinaryOperator("+", other, self)

    def __sub__(self, other):
        return BinaryOperator("-", self, other)

    def __rsub__(self, other):
        return BinaryOperator("-", other, self)

    def __mul__(self, other):
        return BinaryOperator("*", self, other)

    def __rmul__(self, other):
        return BinaryOperator("*", other, self)

    def __truediv__(self, other):
        return BinaryOperator("/", self, other)

    def __floordiv__(self, other):
        return BinaryOperator("//", self, other)

    def __mod__(self, other):
        return BinaryOperator("%", self, other)

    def __matmul__(self, other):
        return BinaryOperator("@", self, other)

    def __rmatmul__(self, other):
        return BinaryOperator("@", other, self)

    def __pow__(self, other):
        return BinaryOperator("**", self, other)

    def __lshift__(self, other):
        return BinaryOperator("<<", self, other)

    def __rshift__(self, other):
        return BinaryOperator(">>", self, other)


class Leaf(Node):
    __slots__ = ()

    def __new__(cls, value):
        return super().__new__(cls, value, precedence["(expressions...)"], tuple())

    def __str__(self):
        return str(self.value)


class UnaryOperator(Node):
    __slots__ = ()

    def __new__(cls, value, child):
        return super().__new__(cls, value, precedence[value + "x"], (child,))

    def __str__(self):
        child = self.children[0]
        operand = str(child)
        if self.precedence >= child.precedence:
            operand = "({})".format(operand)
        return self.value + operand


class BinaryOperator(Node):
    __slots__ = ()

    def __new__(cls, value, lhs, rhs):
        return super().__new__(cls, value, precedence[value], (lhs, rhs))

    def __str__(self):
        lhs, rhs = self.children
        if self.precedence > lhs.precedence:
            lhs = "({})".format(lhs)
        if self.precedence >= rhs.precedence:
            rhs = "({})".format(rhs)
        return "{} {} {}".format(lhs, self.value, rhs)



"""Associative memory implementations.

See :doc:`examples/associative_memory` for an introduction and examples.
"""
import nengo
from nengo.exceptions import ValidationError
from nengo.utils.network import with_self
import numpy as np

from nengo_spa.network import Network
from nengo_spa.networks.selection import IA, Thresholding, WTA
from nengo_spa.vocabulary import VocabularyOrDimParam


class AssociativeMemory(Network):
    """General associative memory network.

    This provides a low-level selection network with the necessary interface
    to include it within the SPA system.

    Parameters
    ----------
    selection_net : Network
        The network that is used to select the response. It needs to accept
        the arguments *n_neurons* (number of neurons to use to represent each
        possible choice) and *n_ensembles* (number of choices). The returned
        network needs to have an *input* attribute to which the utilities for
        each choice are connected and an *output* attribute from which a
        connection will be created to read the selected output(s).
    input_vocab: Vocabulary or int
        The vocabulary to match.
    output_vocab: Vocabulary or int, optional
        The vocabulary to be produced for each match. If
        None, the associative memory will act like an auto-associative memory
        (cleanup memory).
    mapping: dict, str, or sequence of str
        A dictionary that defines the mapping from Semantic Pointers in the
        input vocabulary to Semantic Pointers in the output vocabulary. If set
        to the string ``'by-key'``, the mapping will be done based on the keys
        of the to vocabularies. If a sequence is provided, an auto-associative
        (cleanup) memory with the given set of keys will be created.
    n_neurons : int
        Number of neurons to represent each choice, passed on to the
        *selection_net*.
    label : str, optional
        A name for the ensemble. Used for debugging and visualization.
    seed : int, optional
        The seed used for random number generation.
    add_to_container : bool, optional
        Determines if this Network will be added to the current container.
        See `nengo.Network` for more details.
    vocabs : VocabularyMap, optional
        Maps dimensionalities to the corresponding default vocabularies.
    **selection_net_kwargs : dict
        Additional keyword arguments that will passed to the *selection_net*.
    """

    input_vocab = VocabularyOrDimParam("input_vocab", default=None, readonly=True)
    output_vocab = VocabularyOrDimParam("output_vocab", default=None, readonly=True)

    def __init__(
        self,
        selection_net,
        input_vocab,
        output_vocab=None,
        mapping=None,
        n_neurons=50,
        label=None,
        seed=None,
        add_to_container=None,
        vocabs=None,
        **selection_net_kwargs
    ):
        super(AssociativeMemory, self).__init__(
            label=label, seed=seed, add_to_container=add_to_container, vocabs=vocabs
        )

        if output_vocab is None:
            output_vocab = input_vocab
        elif mapping is None:
            raise ValidationError(
                "The mapping argument needs to be provided if an output "
                "vocabulary is given.",
                attr="mapping",
                obj=self,
            )
        self.input_vocab = input_vocab
        self.output_vocab = output_vocab

        if mapping is None:
            raise TypeError("Must provide 'mapping' argument.")
        elif mapping == "by-key":
            mapping = self.input_vocab.keys()
        elif isinstance(mapping, str):
            raise ValidationError(
                "The mapping argument must be a dictionary, the string "
                "'by-key' or a sequence of strings.",
                attr="mapping",
                obj=self,
            )

        if not hasattr(mapping, "keys"):
            mapping = {k: k for k in mapping}

        if len(mapping) < 1:
            raise ValidationError(
                "At least one item must be provided with the mapping " "argument.",
                attr="mapping",
                obj=self,
            )

        input_keys = mapping.keys()
        input_vectors = [self.input_vocab.parse(key).v for key in input_keys]
        output_keys = [mapping[k] for k in input_keys]
        output_vectors = [self.output_vocab.parse(key).v for key in output_keys]

        input_vectors = np.asarray(input_vectors)
        output_vectors = np.asarray(output_vectors)

        with self:
            self.selection = selection_net(
                n_neurons=n_neurons,
                n_ensembles=len(input_vectors),
                label="selection",
                **selection_net_kwargs
            )
            self.input = nengo.Node(size_in=self.input_vocab.dimensions, label="input")
            self.output = nengo.Node(
                size_in=self.output_vocab.dimensions, label="output"
            )

            nengo.Connection(self.input, self.selection.input, transform=input_vectors)
            nengo.Connection(
                self.selection.output, self.output, transform=output_vectors.T
            )

        self.declare_input(self.input, self.input_vocab)
        self.declare_output(self.output, self.output_vocab)

    @with_self
    def add_default_output(self, key, min_activation_value, n_neurons=50):
        """Adds a Semantic Pointer to output when no other pointer is active.

        Parameters
        ----------
        key : str
            Semantic Pointer to output.
        min_activation_value : float
            Minimum output of another Semantic Pointer to deactivate the
            default output.
        n_neurons : int, optional
            Number of neurons used to represent the default Semantic Pointer.
        """
        assert not hasattr(self, "default_ens"), "Can add default output only once."

        with nengo.presets.ThresholdingEnsembles(0.0):
            setattr(self, "default_ens", nengo.Ensemble(n_neurons, 1, label="default"))
        setattr(self, "bias", nengo.Node(1.0, label="bias"))
        nengo.Connection(self.bias, self.default_ens)
        nengo.Connection(
            self.default_ens,
            self.output,
            transform=np.atleast_2d(self.output_vocab.parse(key).v).T,
        )
        nengo.Connection(
            self.selection.output,
            self.default_ens,
            transform=-np.ones((1, self.selection.output.size_out))
            / min_activation_value,
        )


class IAAssocMem(AssociativeMemory):
    """Associative memory based on the `.IA` network.

    See `AssociativeMemory` and `.IA` for more information.
    """

    def __init__(
        self,
        input_vocab,
        output_vocab=None,
        mapping=None,
        n_neurons=50,
        label=None,
        seed=None,
        add_to_container=None,
        vocabs=None,
        **selection_net_kwargs
    ):
        super(IAAssocMem, self).__init__(
            selection_net=IA,
            input_vocab=input_vocab,
            output_vocab=output_vocab,
            mapping=mapping,
            n_neurons=n_neurons,
            label=label,
            seed=seed,
            add_to_container=add_to_container,
            vocabs=vocabs,
            **selection_net_kwargs
        )
        self.input_reset = self.selection.input_reset
        self.declare_input(self.input_reset, None)


class ThresholdingAssocMem(AssociativeMemory):
    """Associative memory based on `.Thresholding`.

    See `AssociativeMemory` and `.Thresholding` for more information.
    """

    def __init__(
        self,
        threshold,
        input_vocab,
        output_vocab=None,
        mapping=None,
        n_neurons=50,
        label=None,
        seed=None,
        add_to_container=None,
        vocabs=None,
        **selection_net_kwargs
    ):
        selection_net_kwargs["threshold"] = threshold
        super(ThresholdingAssocMem, self).__init__(
            selection_net=Thresholding,
            input_vocab=input_vocab,
            output_vocab=output_vocab,
            mapping=mapping,
            n_neurons=n_neurons,
            label=label,
            seed=seed,
            add_to_container=add_to_container,
            vocabs=vocabs,
            **selection_net_kwargs
        )


class WTAAssocMem(AssociativeMemory):
    """Associative memory based on the `.WTA` network.

    See `AssociativeMemory` and `.WTA` for more information.
    """

    def __init__(
        self,
        threshold,
        input_vocab,
        output_vocab=None,
        mapping=None,
        n_neurons=50,
        label=None,
        seed=None,
        add_to_container=None,
        vocabs=None,
        **selection_net_kwargs
    ):
        selection_net_kwargs["threshold"] = threshold
        super(WTAAssocMem, self).__init__(
            selection_net=WTA,
            input_vocab=input_vocab,
            output_vocab=output_vocab,
            mapping=mapping,
            n_neurons=n_neurons,
            label=label,
            seed=seed,
            add_to_container=add_to_container,
            vocabs=vocabs,
            **selection_net_kwargs
        )



import nengo
import pytest

import nengo_spa as spa
from nengo_spa.algebras.vtb_algebra import VtbAlgebra
from nengo_spa.networks.vtb import VTB
from nengo_spa.testing import assert_sp_close


def test_bind(Simulator, seed, rng):
    vocab = spa.Vocabulary(16, pointer_gen=rng, algebra=VtbAlgebra())
    vocab.populate("A; B")

    with spa.Network(seed=seed) as model:
        vtb = VTB(100, 16)
        nengo.Connection(nengo.Node(vocab["A"].v), vtb.input_left)
        nengo.Connection(nengo.Node(vocab["B"].v), vtb.input_right)
        p = nengo.Probe(vtb.output, synapse=0.03)

    with Simulator(model) as sim:
        sim.run(0.2)

    assert_sp_close(sim.trange(), sim.data[p], vocab.parse("A*B"), skip=0.15, atol=0.3)


@pytest.mark.parametrize("side", ("left", "right"))
def test_unbind(Simulator, side, seed, rng):
    vocab = spa.Vocabulary(36, pointer_gen=rng, algebra=VtbAlgebra())
    vocab.populate("A; B")

    with spa.Network(seed=seed) as model:
        vtb = VTB(100, 36, unbind_left=(side == "left"), unbind_right=(side == "right"))

        if side == "left":
            left = nengo.Node(vocab["B"].v)
            right = nengo.Node(vocab.parse("B*A").v)
        elif side == "right":
            left = nengo.Node(vocab.parse("A*B").v)
            right = nengo.Node(vocab["B"].v)
        else:
            raise ValueError("Invalid 'side' value.")

        nengo.Connection(left, vtb.input_left)
        nengo.Connection(right, vtb.input_right)

        p = nengo.Probe(vtb.output, synapse=0.03)

    with Simulator(model) as sim:
        sim.run(0.2)

    assert_sp_close(
        sim.trange(), sim.data[p], vocab.parse("A * B * ~B"), skip=0.15, atol=0.3
    )



import sys

import nengo
import numpy as np
from numpy.testing import assert_allclose, assert_equal
import pytest

import nengo_spa as spa
from nengo_spa import sym
from nengo_spa.ast.symbolic import FixedScalar, PointerSymbol
from nengo_spa.types import TVocabulary


def test_product_of_scalars(Simulator, seed):
    with spa.Network(seed=seed) as model:
        stimulus = nengo.Node(0.5)
        a = spa.Scalar()
        b = spa.Scalar()
        nengo.Connection(stimulus, a.input)
        nengo.Connection(stimulus, b.input)
        x = a * b
        p = nengo.Probe(x.construct(), synapse=0.03)

    with Simulator(model) as sim:
        sim.run(0.5)

    assert_allclose(sim.data[p][sim.trange() > 0.3], 0.25, atol=0.2)


def test_unary_minus_on_scalar():
    assert (-FixedScalar(1.0)).evaluate() == -1.0


def test_pointer_symbol_network_creation(rng):
    vocab = spa.Vocabulary(16, pointer_gen=rng)
    vocab.populate("A")

    with spa.Network():
        A = PointerSymbol("A", TVocabulary(vocab))
        node = A.construct()
    assert_equal(node.output, vocab["A"].v)


@pytest.mark.parametrize("op", ["-", "~"])
def test_unary_operation_on_pointer_symbol(op, rng):
    vocab = spa.Vocabulary(16, pointer_gen=rng)
    vocab.populate("A")

    with spa.Network():
        x = eval(op + "PointerSymbol('A', TVocabulary(vocab))")
        node = x.construct()
    assert_equal(node.output, vocab.parse(op + "A").v)


@pytest.mark.parametrize("op", ["+", "-", "*"])
def test_binary_operation_on_pointer_symbols(op, rng):
    vocab = spa.Vocabulary(16, pointer_gen=rng)
    vocab.populate("A; B")

    with spa.Network():
        v = TVocabulary(vocab)  # noqa: F841
        x = eval("PointerSymbol('A', v)" + op + "PointerSymbol('B', v)")
        node = x.construct()
    assert_equal(node.output, vocab.parse("A" + op + "B").v)


def test_pointer_symbol_mul_with_array():
    with pytest.raises(TypeError):
        PointerSymbol("X") * np.array([1, 2])


@pytest.mark.parametrize("op", ["+", "-"])
def test_additive_op_fixed_scalar_and_pointer_symbol(op, rng):
    vocab = spa.Vocabulary(16, pointer_gen=rng)
    vocab.populate("A")

    with spa.Network():
        with pytest.raises(TypeError):
            eval("2" + op + "PointerSymbol('A')")


@pytest.mark.parametrize("scalar", [2, np.float64(2)])
def test_multiply_fixed_scalar_and_pointer_symbol(scalar, rng):
    vocab = spa.Vocabulary(16, pointer_gen=rng)
    vocab.populate("A")

    with spa.Network():
        x = scalar * PointerSymbol("A", TVocabulary(vocab))
        node = x.construct()
    assert_equal(node.output, vocab.parse("2 * A").v)


@pytest.mark.parametrize("scalar", [2, np.float64(2)])
def test_divide_pointer_symbol_by_fixed_scalar(scalar, rng):
    vocab = spa.Vocabulary(16, pointer_gen=rng)
    vocab.populate("A")

    with spa.Network():
        x = PointerSymbol("A", TVocabulary(vocab)) / scalar
        node = x.construct()
    assert_equal(node.output, vocab.parse("0.5 * A").v)


def test_fixed_dot(rng):
    vocab = spa.Vocabulary(16, pointer_gen=rng)
    vocab.populate("A; B")

    v = TVocabulary(vocab)
    assert_allclose(
        spa.dot(PointerSymbol("A", v), PointerSymbol("A", v)).evaluate(), 1.0
    )
    assert spa.dot(PointerSymbol("A", v), PointerSymbol("B", v)).evaluate() <= 0.1


@pytest.mark.skipif(sys.version_info < (3, 5), reason="requires Python 3.5")
def test_fixed_dot_matmul(rng):
    vocab = spa.Vocabulary(16, pointer_gen=rng)
    vocab.populate("A; B")

    v = TVocabulary(vocab)  # noqa: F841
    assert_allclose(
        eval("PointerSymbol('A', v) @ PointerSymbol('A', v)").evaluate(), 1.0
    )


def test_translate(rng):
    v1 = spa.Vocabulary(16, pointer_gen=rng)
    v1.populate("A; B")
    v2 = spa.Vocabulary(16, pointer_gen=rng)
    v2.populate("A; B")

    assert_allclose(
        spa.translate(PointerSymbol("A", TVocabulary(v1)), v2).evaluate().dot(v2["A"]),
        1.0,
        atol=0.2,
    )


def test_reinterpret(rng):
    v1 = spa.Vocabulary(16, pointer_gen=rng)
    v1.populate("A; B")
    v2 = spa.Vocabulary(16, pointer_gen=rng)
    v2.populate("A; B")

    assert_equal(
        spa.reinterpret(PointerSymbol("A", TVocabulary(v1)), v2).evaluate().v, v1["A"].v
    )


def test_pointer_symbol_factory():
    ps = sym.A
    assert isinstance(ps, PointerSymbol)
    assert ps.expr == "A"


@pytest.mark.parametrize(
    "ps,expected",
    [
        (sym("A + B * C"), "(A+B*C)"),
        (sym.A + sym.B * sym.C, "A + B * C"),
        (sym("(A + B) * C"), "((A+B)*C)"),
        ((sym.A + sym.B) * sym.C, "(A + B) * C"),
    ],
)
def test_pointer_symbol_factory_expressions(ps, expected):
    assert isinstance(ps, PointerSymbol)
    assert ps.expr == expected



import nengo
from nengo.utils.numpy import rms

import nengo_spa as spa
from nengo_spa.vocabulary import VocabularyMap


def test_basic():
    with spa.Network():
        bind = spa.Superposition(2, vocab=16)

    # all inputs and outputs should share the same vocab
    vocab_a = spa.Network.get_input_vocab(bind.inputs[0])
    vocab_b = spa.Network.get_input_vocab(bind.inputs[1])
    assert vocab_a is vocab_b
    assert vocab_a.dimensions == 16
    assert vocab_b.dimensions == 16


def test_run(Simulator, algebra, seed, rng):
    vocab = spa.Vocabulary(32, pointer_gen=rng, algebra=algebra)
    vocab.populate("A; B")

    with spa.Network(seed=seed, vocabs=VocabularyMap([vocab])) as model:
        model.superpos = spa.Superposition(2, vocab=32)

        def inputA(t):
            if 0 <= t < 0.1:
                return "A"
            else:
                return "B"

        model.input = spa.Transcode(inputA, output_vocab=vocab)
        model.input >> model.superpos.inputs[0]
        spa.sym.A >> model.superpos.inputs[1]

        p = nengo.Probe(model.superpos.output, synapse=0.03)

    with Simulator(model) as sim:
        sim.run(0.2)

    error = rms(vocab.parse("(B+A).normalized()").v - sim.data[p][-1])
    assert error < 0.1

    error = rms(vocab.parse("(A+A).normalized()").v - sim.data[p][100])
    assert error < 0.2



import weakref

import nengo
import numpy as np

from nengo_spa.ast.base import Node, infer_types, Fixed
from nengo_spa.ast.dynamic import ModuleOutput
from nengo_spa.ast.symbolic import FixedScalar
from nengo_spa.exceptions import SpaTypeError
from nengo_spa.typechecks import is_number
from nengo_spa.types import TScalar, TVocabulary


def as_ast_node(obj):
    if isinstance(obj, Node):
        return obj
    elif is_number(obj):
        return FixedScalar(obj)
    elif isinstance(obj, nengo.Network) and hasattr(obj, "output"):
        output = obj.output
    else:
        output = obj

    try:
        # Trying to create weakref on access of weak dict can raise TypeError
        vocab = output_vocab_registry[output]
    except (KeyError, TypeError):
        if getattr(output, "size_out", 0) == 1:
            return ModuleOutput(output, TScalar)
        err = SpaTypeError("{} was not registered as a SPA output.".format(output))
        err.__suppress_context__ = True
        raise err
    finally:
        err = None  # prevent cyclic reference, traceback might reference this

    if vocab is None:
        return ModuleOutput(output, TScalar)
    else:
        return ModuleOutput(output, TVocabulary(vocab))


def as_sink(obj):
    if isinstance(obj, nengo.Network) and hasattr(obj, "input"):
        input_ = obj.input
    else:
        input_ = obj

    try:
        # Trying to create weakref on access of weak dict can raise TypeError
        vocab = input_vocab_registry[input_]
    except (KeyError, TypeError):
        err = SpaTypeError("{} was not registered as a SPA input.".format(input_))
        err.__suppress_context__ = True
        raise err
    finally:
        err = None  # prevent cyclic reference, traceback might reference this

    if vocab is None:
        return ModuleInput(input_, TScalar)
    else:
        return ModuleInput(input_, TVocabulary(vocab))


class ModuleInput:
    """Represents the input to a module with type information.

    Supports the ``>>`` operator to provide input from an AST node. It will
    create a simple connection by default, but a `.RoutedConnection` instance
    if used within the context of an `.ActionSelection` instance.

    Parameters
    ----------
    input_ : NengoObject
        Nengo object that retrieves the module input.
    type_ : nengo_spa.types.Type
        Type of the input.
    """

    routed_mode = False

    def __init__(self, input_, type_):
        self.input = input_
        self.type = type_

    def __rrshift__(self, other):
        if not isinstance(other, Node):
            return NotImplemented
        if self.routed_mode:
            return RoutedConnection(other, self)
        else:
            infer_types(self, other)
            other.connect_to(self.input)


class RoutedConnection:
    """Represents a routed connection from an AST node to a `.ModuleInput`.

    A routed connection is passed through an inhibitable channel to allow the
    routed connection to be disabled.

    Parameters
    ----------
    source : nengo_spa.ast.base.Node
        The AST node providing the source signal.
    sink : ModuleInput
        The module input that the source signal should be routed to.
    """

    free_floating = set()

    def __init__(self, source, sink):
        self.type = infer_types(source, sink)
        self.source = source
        self.sink = sink
        RoutedConnection.free_floating.add(self)

    def connect_to(self, sink):
        return self.source.connect_to(sink)

    def construct(self):
        return self.connect_to(self.sink.input)

    @property
    def fixed(self):
        """Whether the source provides a fixed value."""
        return isinstance(self.source, Fixed)

    def transform(self):
        """For a fixed source, returns the transform to implement the output."""
        assert self.fixed
        if self.type == TScalar:
            return self.source.evaluate()
        else:
            return np.atleast_2d(self.source.evaluate().v).T


class SpaOperatorMixin:
    """Mixin class that implements the SPA operators.

    All operands will be converted to AST node and the implementation of the
    operator itself is delegated to the implementation provided by those nodes.
    """

    @staticmethod
    def __define_unary_op(op):
        def op_impl(self):
            return getattr(as_ast_node(self), op)()

        return op_impl

    @staticmethod
    def __define_binary_op(op):
        def op_impl(self, other):
            return getattr(as_ast_node(self), op)(as_ast_node(other))

        return op_impl

    __invert__ = __define_unary_op.__func__("__invert__")
    __neg__ = __define_unary_op.__func__("__neg__")

    __add__ = __define_binary_op.__func__("__add__")
    __radd__ = __define_binary_op.__func__("__radd__")
    __sub__ = __define_binary_op.__func__("__sub__")
    __rsub__ = __define_binary_op.__func__("__rsub__")
    __mul__ = __define_binary_op.__func__("__mul__")
    __rmul__ = __define_binary_op.__func__("__rmul__")
    __truediv__ = __define_binary_op.__func__("__truediv__")
    __matmul__ = __define_binary_op.__func__("__matmul__")
    __rmatmul__ = __define_binary_op.__func__("__rmatmul__")

    def __rshift__(self, other):
        return as_ast_node(self) >> as_sink(other)

    def __rrshift__(self, other):
        return as_ast_node(other) >> as_sink(self)

    dot = __define_binary_op.__func__("dot")
    rdot = __define_binary_op.__func__("rdot")

    def reinterpret(self, vocab=None):
        return as_ast_node(self).reinterpret(vocab)

    def translate(self, vocab, populate=None, keys=None, solver=None):
        return as_ast_node(self).translate(vocab, populate, keys, solver)


class ConnectorRegistry:
    """Registry associating connectors with vocabularies and enable SPA syntax.

    A connector is either an input or output to a SPA module.

    Declaring a Nengo object as connector dynamically changes the class of the
    instance to inject the `.SpaOperatorMixin` to allow the object to be used
    in SPA rules. These substituted classes will be cached to use the same
    type instance where appropriate (i.e. two Nengo objects of identical type A
    will still be of identical type B after being declared as connectors and
    B inherits from A and *SpaOperatorMixin*).
    """

    _type_cache = {}

    def __init__(self):
        self._registry = weakref.WeakKeyDictionary()

    def __contains__(self, key):
        return key in self._registry

    def __getitem__(self, key):
        return self._registry[key]

    def declare_connector(self, obj, vocab):
        """Declares a connector.

        Parameters
        ----------
        obj : nengo.base.NengoObject
            Nengo object to use as a connector.
        vocab: Vocabulary
            Vocabulary to assign to the connector.
        """
        try:
            extended_type = self._type_cache[obj.__class__]
        except KeyError:
            extended_type = type(
                "Connector<%s>" % (obj.__class__.__name__),
                (obj.__class__, SpaOperatorMixin),
                {},
            )
            self._type_cache[obj.__class__] = extended_type
        obj.__class__ = extended_type
        self._registry[obj] = vocab
        return obj


input_vocab_registry = ConnectorRegistry()
output_vocab_registry = ConnectorRegistry()



import nengo
from nengo.exceptions import ValidationError
import numpy as np

from nengo_spa.algebras.base import AbstractAlgebra
from nengo_spa.algebras.hrr_algebra import HrrAlgebra
from nengo_spa.ast.base import Fixed, infer_types, TypeCheckedBinaryOp
from nengo_spa.typechecks import is_array, is_array_like, is_number
from nengo_spa.types import TAnyVocab, TScalar, TVocabulary


class SemanticPointer(Fixed):
    """A Semantic Pointer, based on Holographic Reduced Representations.

    Operators are overloaded so that ``+`` and ``-`` are addition,
    ``*`` is circular convolution, and ``~`` is the inversion operator.

    Parameters
    ----------
    data : array_like
        The vector constituting the Semantic Pointer.
    vocab : Vocabulary, optional
        Vocabulary that the Semantic Pointer is considered to be part of.
        Mutually exclusive with the *algebra* argument.
    algebra : AbstractAlgebra, optional
        Algebra used to perform vector symbolic operations on the Semantic
        Pointer. Defaults to `.CircularConvolutionAlgebra`. Mutually exclusive
        with the *vocab* argument.
    name : str, optional
        A name for the Semantic Pointer.

    Attributes
    ----------
    v : array_like
        The vector constituting the Semantic Pointer.
    algebra : AbstractAlgebra
        Algebra that defines the vector symbolic operations on this Semantic
        Pointer.
    vocab : Vocabulary or None
        The vocabulary the this Semantic Pointer is considered to be part of.
    name : str or None
        Name of the Semantic Pointer.
    """

    MAX_NAME = 1024

    def __init__(self, data, vocab=None, algebra=None, name=None):
        super(SemanticPointer, self).__init__(
            TAnyVocab if vocab is None else TVocabulary(vocab)
        )
        self.algebra = self._get_algebra(vocab, algebra)

        self.v = np.array(data, dtype=float)
        if len(self.v.shape) != 1:
            raise ValidationError("'data' must be a vector", "data", self)
        self.v.setflags(write=False)

        self.vocab = vocab

        if name is not None and len(name) > self.MAX_NAME:
            name = "%s..." % (name[: self.MAX_NAME - 3])
        self.name = name

    def _get_algebra(cls, vocab, algebra):
        if algebra is None:
            if vocab is None:
                algebra = HrrAlgebra()
            else:
                algebra = vocab.algebra
        elif vocab is not None and vocab.algebra is not algebra:
            raise ValueError("vocab and algebra argument are mutually exclusive")
        if not isinstance(algebra, AbstractAlgebra):
            raise ValidationError(
                "'algebra' must be an instance of AbstractAlgebra", "algebra", algebra
            )
        return algebra

    def _get_unary_name(self, op):
        return "{}({})".format(op, self.name) if self.name else None

    def _get_method_name(self, method):
        return "({}).{}()".format(self.name, method) if self.name else None

    def _get_binary_name(self, other, op, swap=False):
        if isinstance(other, SemanticPointer):
            other_name = other.name
        else:
            other_name = str(other)
        self_name = self.name
        if self_name and other_name:
            if swap:
                self_name, other_name = other_name, self.name
            return "({}){}({})".format(self_name, op, other_name)
        else:
            return None

    def evaluate(self):
        return self

    def connect_to(self, sink, **kwargs):
        return nengo.Connection(self.construct(), sink, **kwargs)

    def construct(self):
        return nengo.Node(self.v, label=str(self).format(len(self)))

    def normalized(self):
        """Normalize the Semantic Pointer and return it as a new object.

        If the vector length is zero, the Semantic Pointer will be returned
        unchanged.

        The original object is not modified.
        """
        nrm = np.linalg.norm(self.v)
        if nrm <= 0.0:
            nrm = 1.0
        return SemanticPointer(
            self.v / nrm,
            vocab=self.vocab,
            algebra=self.algebra,
            name=self._get_method_name("normalized"),
        )

    def unitary(self):
        """Make the Semantic Pointer unitary and return it as a new object.

        The original object is not modified.

        A unitary Semantic Pointer has the property that it does not change
        the length of Semantic Pointers it is bound with using circular
        convolution.
        """
        return SemanticPointer(
            self.algebra.make_unitary(self.v),
            vocab=self.vocab,
            algebra=self.algebra,
            name=self._get_method_name("unitary"),
        )

    def copy(self):
        """Return another semantic pointer with the same data."""
        return SemanticPointer(
            data=self.v, vocab=self.vocab, algebra=self.algebra, name=self.name
        )

    def length(self):
        """Return the L2 norm of the vector."""
        return np.linalg.norm(self.v)

    def __len__(self):
        """Return the number of dimensions in the vector."""
        return len(self.v)

    def __str__(self):
        if self.name:
            return "SemanticPointer<{}>".format(self.name)
        else:
            return repr(self)

    def __repr__(self):
        return "SemanticPointer({!r}, vocab={!r}, algebra={!r}, name={!r}".format(
            self.v, self.vocab, self.algebra, self.name
        )

    @TypeCheckedBinaryOp(Fixed)
    def __add__(self, other):
        return self._add(other, swap=False)

    @TypeCheckedBinaryOp(Fixed)
    def __radd__(self, other):
        return self._add(other, swap=True)

    def _add(self, other, swap=False):
        type_ = infer_types(self, other)
        vocab = None if type_ == TAnyVocab else type_.vocab
        if vocab is None:
            self._ensure_algebra_match(other)
        other_pointer = other.evaluate()
        a, b = self.v, other_pointer.v
        if swap:
            a, b = b, a
        return SemanticPointer(
            data=self.algebra.superpose(a, b),
            vocab=vocab,
            algebra=self.algebra,
            name=self._get_binary_name(other_pointer, "+", swap),
        )

    def __neg__(self):
        return SemanticPointer(
            data=-self.v,
            vocab=self.vocab,
            algebra=self.algebra,
            name=self._get_unary_name("-"),
        )

    def __sub__(self, other):
        return self + (-other)

    def __rsub__(self, other):
        return (-self) + other

    def __mul__(self, other):
        """Multiplication of two SemanticPointers is circular convolution.

        If multiplied by a scalar, we do normal multiplication.
        """
        return self._mul(other, swap=False)

    def __rmul__(self, other):
        """Multiplication of two SemanticPointers is circular convolution.

        If multiplied by a scalar, we do normal multiplication.
        """
        return self._mul(other, swap=True)

    def _mul(self, other, swap=False):
        if is_number(other):
            return SemanticPointer(
                data=self.v * other,
                vocab=self.vocab,
                algebra=self.algebra,
                name=self._get_binary_name(other, "*", swap),
            )
        elif is_array(other):
            raise TypeError(
                "Multiplication of Semantic Pointers with arrays in not allowed."
            )
        elif isinstance(other, Fixed):
            if other.type == TScalar:
                return SemanticPointer(
                    data=self.v * other.evaluate(),
                    vocab=self.vocab,
                    algebra=self.algebra,
                    name=self._get_binary_name(other, "*", swap),
                )
            else:
                return self._bind(other, swap=swap)
        else:
            return NotImplemented

    def __truediv__(self, other):
        if is_number(other):
            if other == 0:
                raise ZeroDivisionError("Semantic Pointer division by zero")
            return SemanticPointer(
                data=self.v / other,
                vocab=self.vocab,
                algebra=self.algebra,
                name=self._get_binary_name(other, "/"),
            )
        elif is_array(other):
            raise TypeError("Division of Semantic Pointers with arrays is not allowed.")
        else:
            return NotImplemented

    def __invert__(self):
        """Return a reorganized vector that acts as an inverse for convolution.

        This reorganization turns circular convolution into circular
        correlation, meaning that ``A*B*~B`` is approximately ``A``.

        For the vector ``[1, 2, 3, 4, 5]``, the inverse is ``[1, 5, 4, 3, 2]``.
        """
        return SemanticPointer(
            data=self.algebra.invert(self.v),
            vocab=self.vocab,
            algebra=self.algebra,
            name=self._get_unary_name("~"),
        )

    def bind(self, other):
        """Return the binding of two SemanticPointers."""
        return self._bind(other, swap=False)

    def rbind(self, other):
        """Return the binding of two SemanticPointers."""
        return self._bind(other, swap=True)

    def _bind(self, other, swap=False):
        type_ = infer_types(self, other)
        vocab = None if type_ == TAnyVocab else type_.vocab
        if vocab is None:
            self._ensure_algebra_match(other)
        other_pointer = other.evaluate()
        a, b = self.v, other_pointer.v
        if swap:
            a, b = b, a
        return SemanticPointer(
            data=self.algebra.bind(a, b),
            vocab=vocab,
            algebra=self.algebra,
            name=self._get_binary_name(other_pointer, "*", swap),
        )

    def get_binding_matrix(self, swap_inputs=False):
        """Return the matrix that does a binding with this vector.

        This should be such that
        ``A*B == dot(A.get_binding_matrix(), B.v)``.
        """
        return self.algebra.get_binding_matrix(self.v, swap_inputs=swap_inputs)

    def dot(self, other):
        """Return the dot product of the two vectors."""
        if isinstance(other, Fixed):
            infer_types(self, other)
            other = other.evaluate().v
        if is_array_like(other):
            return np.dot(self.v, other)
        else:
            return other.dot(self)

    def __matmul__(self, other):
        return self.dot(other)

    def compare(self, other):
        """Return the similarity between two SemanticPointers.

        This is the normalized dot product, or (equivalently), the cosine of
        the angle between the two vectors.
        """
        if isinstance(other, SemanticPointer):
            infer_types(self, other)
            other = other.evaluate().v
        scale = np.linalg.norm(self.v) * np.linalg.norm(other)
        if scale == 0:
            return 0
        return np.dot(self.v, other) / scale

    def reinterpret(self, vocab):
        """Reinterpret the Semantic Pointer as part of vocabulary *vocab*.

        The *vocab* parameter can be set to *None* to clear the associated
        vocabulary and allow the *source* to be interpreted as part of the
        vocabulary of any Semantic Pointer it is combined with.
        """
        return SemanticPointer(self.v, vocab=vocab, name=self.name)

    def translate(self, vocab, populate=None, keys=None, solver=None):
        """Translate the Semantic Pointer to vocabulary *vocab*.

        The translation of a Semantic Pointer uses some form of projection to
        convert the Semantic Pointer to a Semantic Pointer of another
        vocabulary. By default the outer products of terms in the source and
        target vocabulary are used, but if *solver* is given, it is used to
        find a least squares solution for this projection.

        Parameters
        ----------
        vocab : Vocabulary
            Target vocabulary.
        populate : bool, optional
            Whether the target vocabulary should be populated with missing
            keys.  This is done by default, but with a warning. Set this
            explicitly to *True* or *False* to silence the warning or raise an
            error.
        keys : list, optional
            All keys to translate. If *None*, all keys in the source vocabulary
            will be translated.
        solver : nengo.Solver, optional
            If given, the solver will be used to solve the least squares
            problem to provide a better projection for the translation.
        """
        tr = self.vocab.transform_to(vocab, populate=populate, keys=keys, solver=solver)
        return SemanticPointer(
            np.dot(tr, self.evaluate().v), vocab=vocab, name=self.name
        )

    def distance(self, other):
        """Return a distance measure between the vectors.

        This is ``1-cos(angle)``, so that it is 0 when they are identical, and
        the distance gets larger as the vectors are farther apart.
        """
        return 1 - self.compare(other)

    def mse(self, other):
        """Return the mean-squared-error between two vectors."""
        if isinstance(other, SemanticPointer):
            infer_types(self, other)
            other = other.evaluate().v
        return np.sum((self.v - other) ** 2) / len(self.v)

    def _ensure_algebra_match(self, other):
        """Check the algebra of the *other*.

        If the *other* parameter is a `SemanticPointer` and uses a different
        algebra, a `TypeError` will be raised.
        """
        if isinstance(other, SemanticPointer):
            if self.algebra is not other.algebra:
                raise TypeError(
                    "Operation not supported for SemanticPointer with "
                    "different algebra."
                )


class Identity(SemanticPointer):
    """Identity element.

    Parameters
    ----------
    n_dimensions : int
        Dimensionality of the identity vector.
    vocab : Vocabulary, optional
        Vocabulary that the Semantic Pointer is considered to be part of.
        Mutually exclusive with the *algebra* argument.
    algebra : AbstractAlgebra, optional
        Algebra used to perform vector symbolic operations on the Semantic
        Pointer. Defaults to `.CircularConvolutionAlgebra`. Mutually exclusive
        with the *vocab* argument.
    """

    def __init__(self, n_dimensions, vocab=None, algebra=None):
        data = self._get_algebra(vocab, algebra).identity_element(n_dimensions)
        super(Identity, self).__init__(
            data, vocab=vocab, algebra=algebra, name="Identity"
        )


class AbsorbingElement(SemanticPointer):
    r"""Absorbing element.

    If :math:`z` denotes the absorbing element, :math:`v \circledast z = c z`,
    where :math:`v` is a Semantic Pointer and :math:`c` is a real-valued
    scalar. Furthermore :math:`\|z\| = 1`.

    Parameters
    ----------
    n_dimensions : int
        Dimensionality of the identity vector.
    vocab : Vocabulary, optional
        Vocabulary that the Semantic Pointer is considered to be part of.
        Mutually exclusive with the *algebra* argument.
    algebra : AbstractAlgebra, optional
        Algebra used to perform vector symbolic operations on the Semantic
        Pointer. Defaults to `.CircularConvolutionAlgebra`. Mutually exclusive
        with the *vocab* argument.
    """

    def __init__(self, n_dimensions, vocab=None, algebra=None):
        data = self._get_algebra(vocab, algebra).absorbing_element(n_dimensions)
        super(AbsorbingElement, self).__init__(
            data, vocab=vocab, algebra=algebra, name="AbsorbingElement"
        )


class Zero(SemanticPointer):
    """Zero element.

    Parameters
    ----------
    n_dimensions : int
        Dimensionality of the identity vector.
    vocab : Vocabulary, optional
        Vocabulary that the Semantic Pointer is considered to be part of.
        Mutually exclusive with the *algebra* argument.
    algebra : AbstractAlgebra, optional
        Algebra used to perform vector symbolic operations on the Semantic
        Pointer. Defaults to `.CircularConvolutionAlgebra`. Mutually exclusive
        with the *vocab* argument.
    """

    def __init__(self, n_dimensions, vocab=None, algebra=None):
        data = self._get_algebra(vocab, algebra).zero_element(n_dimensions)
        super(Zero, self).__init__(data, vocab=vocab, algebra=algebra, name="Zero")



import nengo
from nengo.utils.numpy import rms
import pytest

import nengo_spa as spa
from nengo_spa.testing import assert_sp_close


def test_basic():
    with spa.Network():
        bind = spa.Bind(vocab=16)

    # all inputs and outputs should share the same vocab
    vocab_a = spa.Network.get_input_vocab(bind.input_left)
    vocab_b = spa.Network.get_input_vocab(bind.input_right)
    assert vocab_a is vocab_b
    assert vocab_a.dimensions == 16
    assert vocab_b.dimensions == 16


def test_run(Simulator, algebra, seed, rng):
    vocab = spa.Vocabulary(16, pointer_gen=rng, algebra=algebra)
    vocab.populate("A; B")

    with spa.Network(seed=seed) as model:
        model.bind = spa.Bind(vocab)

        def inputA(t):
            if 0 <= t < 0.1:
                return "A"
            else:
                return "B"

        model.input = spa.Transcode(inputA, output_vocab=vocab)
        model.input >> model.bind.input_left
        spa.sym.A >> model.bind.input_right

        p = nengo.Probe(model.bind.output, synapse=0.03)

    with Simulator(model) as sim:
        sim.run(0.2)

    error = rms(vocab.parse("(B*A).normalized()").v - sim.data[p][-1])
    assert error < 0.15

    error = rms(vocab.parse("(A*A).normalized()").v - sim.data[p][100])
    assert error < 0.15


@pytest.mark.parametrize("side", ("left", "right"))
def test_unbind(Simulator, algebra, side, seed, rng):
    vocab = spa.Vocabulary(64, pointer_gen=rng, algebra=algebra)
    vocab.populate("A; B")

    with spa.Network(seed=seed) as model:
        model.bind = spa.Bind(
            vocab, unbind_left=(side == "left"), unbind_right=(side == "right")
        )

        if side == "left":
            spa.sym.B >> model.bind.input_left
            spa.sym.B * spa.sym.A >> model.bind.input_right
        elif side == "right":
            spa.sym.A * spa.sym.B >> model.bind.input_left
            spa.sym.B >> model.bind.input_right
        else:
            raise ValueError("Invalid 'side' value.")

        p = nengo.Probe(model.bind.output, synapse=0.03)

    with Simulator(model) as sim:
        sim.run(0.2)

    assert_sp_close(
        sim.trange(), sim.data[p], vocab.parse("A * B * ~B"), skip=0.15, atol=0.3
    )



from nengo_spa import version

from nengo_spa.action_selection import ActionSelection
from nengo_spa.ast.symbolic import sym
from nengo_spa.operators import dot, reinterpret, translate
from nengo_spa.examine import pairs, similarity, text
from nengo_spa.modules import (
    AssociativeMemory,
    IAAssocMem,
    ThresholdingAssocMem,
    WTAAssocMem,
    BasalGanglia,
    Bind,
    Compare,
    Product,
    Scalar,
    State,
    Superposition,
    Thalamus,
    Transcode,
)
from nengo_spa.network import create_inhibit_node, ifmax, Network
from nengo_spa.semantic_pointer import SemanticPointer
from nengo_spa.vector_generation import (
    AxisAlignedVectors,
    ExpectedUnitLengthVectors,
    OrthonormalVectors,
    UnitaryVectors,
    UnitLengthVectors,
)
from nengo_spa.vocabulary import Vocabulary


__copyright__ = "2013-2020, Applied Brain Research"
__license__ = "Free for non-commercial use; see LICENSE.rst"
__version__ = version.version



import numpy as np

from nengo_spa.algebras.vtb_algebra import VtbAlgebra
from nengo_spa.semantic_pointer import SemanticPointer
from nengo_spa.vector_generation import UnitLengthVectors


def test_is_singleton():
    assert VtbAlgebra() is VtbAlgebra()


def test_is_valid_dimensionality():
    assert not VtbAlgebra().is_valid_dimensionality(-1)
    assert not VtbAlgebra().is_valid_dimensionality(0)
    assert not VtbAlgebra().is_valid_dimensionality(15)
    assert not VtbAlgebra().is_valid_dimensionality(24)
    assert VtbAlgebra().is_valid_dimensionality(1)
    assert VtbAlgebra().is_valid_dimensionality(16)
    assert VtbAlgebra().is_valid_dimensionality(25)


def test_get_swapping_matrix(rng):
    gen = UnitLengthVectors(64, rng=rng)
    a = SemanticPointer(next(gen), algebra=VtbAlgebra()).v
    b = SemanticPointer(next(gen), algebra=VtbAlgebra()).v

    m = VtbAlgebra().get_swapping_matrix(64)
    assert np.allclose(VtbAlgebra().bind(a, b), np.dot(m, VtbAlgebra().bind(b, a)))



from nengo.params import Default, IntParam

from nengo_spa.network import Network
from nengo_spa.vocabulary import VocabularyOrDimParam


class Superposition(Network):
    """Network for superposing multiple inputs.

    Parameters
    ----------
    n_inputs : int
        Number of inputs.
    vocab : Vocabulary or int
        The vocabulary to use to interpret the vector. If an integer is given,
        the default vocabulary of that dimensionality will be used.
    neurons_per_dimension : int, optional (Default: 200)
        Number of neurons to use in each dimension.
    **kwargs : dict
        Keyword arguments passed through to `nengo_spa.Network`.

    Attributes
    ----------
    inputs : sequence
        Inputs.
    output : nengo.Node
        Output.
    """

    vocab = VocabularyOrDimParam("vocab", default=None, readonly=True)
    neurons_per_dimension = IntParam(
        "neurons_per_dimension", default=200, low=1, readonly=True
    )
    n_inputs = IntParam("n_inputs", optional=False, low=1, readonly=True)

    def __init__(
        self, n_inputs, vocab=Default, neurons_per_dimension=Default, **kwargs
    ):
        super(Superposition, self).__init__(**kwargs)

        self.vocab = vocab
        self.neurons_per_dimension = neurons_per_dimension
        self.n_inputs = n_inputs

        with self:
            (
                self.superposition_net,
                self.inputs,
                self.output,
            ) = self.vocab.algebra.implement_superposition(
                self.neurons_per_dimension, self.vocab.dimensions, self.n_inputs
            )

        for inp in self.inputs:
            self.declare_input(inp, self.vocab)
        self.declare_output(self.output, self.vocab)



import numpy as np
import pytest

from nengo_spa.algebras.base import AbstractAlgebra
from nengo_spa.vector_generation import UnitLengthVectors


@pytest.mark.parametrize("d", [16, 25])
def test_make_unitary(algebra, d, rng):
    a = next(UnitLengthVectors(d, rng=rng))
    b = algebra.make_unitary(a)
    for i in range(3):
        assert np.allclose(1, np.linalg.norm(a))
        assert np.allclose(1, np.linalg.norm(b))
        a = algebra.bind(a, b)
        assert np.allclose(1, np.linalg.norm(a))


def test_superpose(algebra, rng):
    gen = UnitLengthVectors(16, rng=rng)
    a = next(gen)
    b = next(gen)
    # Orthogonalize
    b -= np.dot(a, b) * a
    b /= np.linalg.norm(b)

    r = algebra.superpose(a, b)
    for v in (a, b):
        assert np.dot(v, r / np.linalg.norm(r)) > 0.6


@pytest.mark.parametrize("d", [25, 36])
def test_binding_and_invert(algebra, d, rng):
    dissimilarity_passed = 0
    unbinding_passed = 0
    for i in range(10):
        gen = UnitLengthVectors(d, rng=rng)
        a = next(gen)
        b = next(gen)
        bound = algebra.bind(a, b)
        r = algebra.bind(bound, algebra.invert(b))
        for v in (a, b):
            dissimilarity_passed += np.dot(v, bound / np.linalg.norm(b)) < 0.7
        unbinding_passed += np.dot(a, r / np.linalg.norm(r)) > 0.6
    assert dissimilarity_passed >= 2 * 8
    assert unbinding_passed >= 8


def test_dimensionality_mismatch_exception(algebra):
    with pytest.raises(ValueError):
        algebra.bind(np.ones(16), np.ones(25))
    with pytest.raises(ValueError):
        algebra.superpose(np.ones(16), np.ones(25))


def test_get_binding_matrix(algebra, rng):
    gen = UnitLengthVectors(16, rng=rng)
    a = next(gen)
    b = next(gen)

    m = algebra.get_binding_matrix(b)

    assert np.allclose(algebra.bind(a, b), np.dot(m, a))


def test_get_inversion_matrix(algebra, rng):
    a = next(UnitLengthVectors(16, rng=rng))
    m = algebra.get_inversion_matrix(16)
    assert np.allclose(algebra.invert(a), np.dot(m, a))


def test_absorbing_element(algebra, rng):
    a = next(UnitLengthVectors(16, rng=rng))
    try:
        p = algebra.absorbing_element(16)
        r = algebra.bind(a, p)
        r /= np.linalg.norm(r)
        assert np.allclose(p, r) or np.allclose(p, -r)
    except NotImplementedError:
        pass


def test_identity_element(algebra, rng):
    a = next(UnitLengthVectors(16, rng=rng))
    p = algebra.identity_element(16)
    assert np.allclose(algebra.bind(a, p), a)


def test_zero_element(algebra, rng):
    a = next(UnitLengthVectors(16, rng=rng))
    p = algebra.zero_element(16)
    assert np.all(p == 0.0)
    assert np.allclose(algebra.bind(a, p), 0.0)


def test_isinstance_check(algebra):
    assert isinstance(algebra, AbstractAlgebra)


class DummyAlgebra:
    def is_valid_dimensionality(self, d):
        pass

    def make_unitary(self, v):
        pass

    def superpose(self, a, b):
        pass

    def bind(self, a, b):
        pass

    def invert(self, v):
        pass

    def get_binding_matrix(self, v, swap_inputs=False):
        pass

    def get_inversion_matrix(self, d):
        pass

    def implement_superposition(self, n_neurons_per_d, d, n):
        pass

    def implement_binding(self, n_neurons_per_d, d, unbind_left, unbind_right):
        pass

    def absorbing_element(self, d):
        pass

    def identity_element(self, d):
        pass

    def zero_element(self, d):
        pass


@pytest.mark.filterwarnings("ignore:.*do not rely on pure duck-typing")
def test_isinstance_ducktyping_check():
    assert isinstance(DummyAlgebra(), AbstractAlgebra)



import nengo
import pytest

import nengo_spa as spa
from nengo_spa.connectors import as_ast_node
from nengo_spa.exceptions import SpaTypeError


@pytest.mark.parametrize(
    "construct_obj",
    (
        lambda: nengo.Node([1.0, 2.0]),
        lambda: [None],
    ),
)
def test_as_ast_node_error(construct_obj):
    with spa.Network():
        with pytest.raises(SpaTypeError):
            as_ast_node(construct_obj())



from collections.abc import Iterable

import numpy as np


def is_array(obj):
    # np.generic allows us to return true for scalars as well as true arrays
    return isinstance(obj, (np.ndarray, np.generic))


def is_array_like(obj):
    # While it's possible that there are some iterables other than list/tuple
    # that can be made into arrays, it's very likely that those arrays
    # will have dtype=object, which is likely to cause unexpected issues.
    return is_array(obj) or is_number(obj) or isinstance(obj, (list, tuple))


def is_float(obj):
    return isinstance(obj, (float, np.float))


def is_integer(obj):
    return isinstance(obj, (int, np.integer))


def is_iterable(obj):
    if isinstance(obj, np.ndarray):
        return obj.ndim > 0
    else:
        return isinstance(obj, Iterable)


def is_number(obj):
    return is_integer(obj) or is_float(obj)



import sys

import nengo
from nengo.exceptions import NengoWarning, ValidationError
import numpy as np
from numpy.testing import assert_equal
import pytest

import nengo_spa as spa
from nengo_spa.algebras.base import AbstractAlgebra
from nengo_spa.algebras.hrr_algebra import HrrAlgebra
from nengo_spa.ast.symbolic import PointerSymbol
from nengo_spa.exceptions import SpaTypeError
from nengo_spa.semantic_pointer import AbsorbingElement, Identity, SemanticPointer, Zero
from nengo_spa.vector_generation import UnitLengthVectors
from nengo_spa.testing import assert_sp_close
from nengo_spa.types import TVocabulary
from nengo_spa.vocabulary import Vocabulary


def test_init():
    a = SemanticPointer([1, 2, 3, 4])
    assert len(a) == 4

    a = SemanticPointer([1, 2, 3, 4, 5])
    assert len(a) == 5

    a = SemanticPointer(list(range(100)))
    assert len(a) == 100

    with pytest.raises(ValidationError):
        a = SemanticPointer(np.zeros((2, 2)))

    with pytest.raises(ValidationError):
        a = SemanticPointer(-1)
    with pytest.raises(ValidationError):
        a = SemanticPointer(0)
    with pytest.raises(ValidationError):
        a = SemanticPointer(1.7)
    with pytest.raises(ValidationError):
        a = SemanticPointer(None)
    with pytest.raises(TypeError):
        a = SemanticPointer(int)


def test_length(rng):
    a = SemanticPointer([1, 1])
    assert np.allclose(a.length(), np.sqrt(2))
    a = SemanticPointer(next(UnitLengthVectors(10, rng=rng))) * 1.2
    assert np.allclose(a.length(), 1.2)


def test_normalized():
    a = SemanticPointer([1, 1]).normalized()
    b = a.normalized()
    assert a is not b
    assert np.allclose(b.length(), 1)


def test_normalized_zero():
    a = SemanticPointer([0, 0]).normalized()
    assert np.allclose(a.v, [0, 0])


@pytest.mark.parametrize("d", [64, 65, 100])
def test_make_unitary(algebra, d, rng):
    if not algebra.is_valid_dimensionality(d):
        return

    a = SemanticPointer(next(UnitLengthVectors(d, rng=rng)), algebra=algebra)
    b = a.unitary()
    assert a is not b
    assert np.allclose(1, b.length())
    assert np.allclose(1, (b * b).length())
    assert np.allclose(1, (b * b * b).length())


def test_add_sub(algebra, rng):
    gen = UnitLengthVectors(10, rng=rng)
    a = SemanticPointer(next(gen), algebra=algebra)
    b = SemanticPointer(next(gen), algebra=algebra)
    c = a.copy()
    d = b.copy()

    c += b
    d -= -a

    assert np.allclose((a + b).v, algebra.superpose(a.v, b.v))
    assert np.allclose((a + b).v, c.v)
    assert np.allclose((a + b).v, d.v)
    assert np.allclose((a + b).v, (a - (-b)).v)


@pytest.mark.parametrize("d", [64, 65])
def test_binding_and_inversion(algebra, d, rng):
    if not algebra.is_valid_dimensionality(d):
        return

    gen = UnitLengthVectors(d, rng=rng)

    a = SemanticPointer(next(gen), algebra=algebra)
    b = SemanticPointer(next(gen), algebra=algebra)
    identity = Identity(d, algebra=algebra)

    c = a.copy()
    c *= b

    conv_ans = algebra.bind(a.v, b.v)

    assert np.allclose((a * b).v, conv_ans)
    assert np.allclose(a.bind(b).v, conv_ans)
    assert np.allclose(c.v, conv_ans)
    assert np.allclose((a * identity).v, a.v)
    assert (a * b * ~b).compare(a) > 0.6


def test_multiply(rng):
    a = SemanticPointer(next(UnitLengthVectors(50, rng=rng)))

    assert np.allclose((a * 5).v, a.v * 5)
    assert np.allclose((5 * a).v, a.v * 5)
    assert np.allclose((a * 5.7).v, a.v * 5.7)
    assert np.allclose((5.7 * a).v, a.v * 5.7)
    assert np.allclose((0 * a).v, np.zeros(50))
    assert np.allclose((1 * a).v, a.v)

    numpy_scalar = np.float64(5.7)
    assert np.allclose((numpy_scalar * a).v, a.v * numpy_scalar)
    assert np.allclose((a * numpy_scalar).v, a.v * numpy_scalar)

    with pytest.raises(Exception):
        a * None
    with pytest.raises(Exception):
        a * "string"
    with pytest.raises(TypeError):
        a * np.array([1, 2])


def test_divide(rng):
    a = SemanticPointer(next(UnitLengthVectors(50, rng=rng)))

    assert np.allclose((a / 5).v, a.v / 5)
    assert np.allclose((a / 5.7).v, a.v / 5.7)
    assert np.allclose((a / np.float64(5.7)).v, a.v / 5.7)
    assert np.allclose((a / 1.0).v, a.v)

    with pytest.raises(ZeroDivisionError):
        a / 0
    with pytest.raises(TypeError):
        5 / a
    with pytest.raises(TypeError):
        a / None
    with pytest.raises(TypeError):
        a / np.ones(50)
    with pytest.raises(TypeError):
        a / SemanticPointer(next(UnitLengthVectors(50, rng=rng)))


def test_compare(rng):
    gen = UnitLengthVectors(50, rng=rng)
    a = SemanticPointer(next(gen)) * 10
    b = SemanticPointer(next(gen)) * 0.1

    assert a.compare(a) > 0.99
    assert a.compare(b) < 0.2
    assert np.allclose(a.compare(b), a.dot(b) / (a.length() * b.length()))


def test_dot(rng):
    gen = UnitLengthVectors(50, rng=rng)
    a = SemanticPointer(next(gen)) * 1.1
    b = SemanticPointer(next(gen)) * (-1.5)
    assert np.allclose(a.dot(b), np.dot(a.v, b.v))
    assert np.allclose(a.dot(b.v), np.dot(a.v, b.v))
    assert np.allclose(a.dot(list(b.v)), np.dot(a.v, b.v))
    assert np.allclose(a.dot(tuple(b.v)), np.dot(a.v, b.v))


@pytest.mark.skipif(sys.version_info < (3, 5), reason="requires Python 3.5")
def test_dot_matmul(rng):
    gen = UnitLengthVectors(50, rng=rng)
    a = SemanticPointer(next(gen)) * 1.1
    b = SemanticPointer(next(gen)) * (-1.5)
    assert np.allclose(eval("a @ b"), np.dot(a.v, b.v))


def test_distance(rng):
    gen = UnitLengthVectors(50, rng=rng)
    a = SemanticPointer(next(gen))
    b = SemanticPointer(next(gen))
    assert a.distance(a) < 1e-5
    assert a.distance(b) > 0.7


def test_len():
    a = SemanticPointer(next(UnitLengthVectors(5)))
    assert len(a) == 5

    a = SemanticPointer(list(range(10)))
    assert len(a) == 10


def test_copy():
    a = SemanticPointer(next(UnitLengthVectors(5)))
    b = a.copy()
    assert a is not b
    assert a.v is not b.v
    assert np.allclose(a.v, b.v)
    assert a.algebra is b.algebra
    assert a.vocab is b.vocab
    assert a.name is b.name


def test_mse(rng):
    gen = UnitLengthVectors(50, rng=rng)
    a = SemanticPointer(next(gen))
    b = SemanticPointer(next(gen))

    assert np.allclose(((a - b).length() ** 2) / 50, a.mse(b))


def test_binding_matrix(algebra, rng):
    gen = UnitLengthVectors(64, rng=rng)
    a = SemanticPointer(next(gen), algebra=algebra)
    b = SemanticPointer(next(gen), algebra=algebra)

    m = b.get_binding_matrix()
    m_swapped = a.get_binding_matrix(swap_inputs=True)

    assert np.allclose((a * b).v, np.dot(m, a.v))
    assert np.allclose((a * b).v, np.dot(m_swapped, b.v))


@pytest.mark.parametrize("op", ("~a", "-a", "a+a", "a-a", "a*a", "2*a"))
def test_ops_preserve_vocab(op):
    v = Vocabulary(50)
    a = SemanticPointer(next(UnitLengthVectors(50)), vocab=v)  # noqa: F841
    x = eval(op)
    assert x.vocab is v


@pytest.mark.parametrize("op", ("a+b", "a-b", "a*b", "a.dot(b)", "a.compare(b)"))
def test_ops_check_vocab_compatibility(op):
    gen = UnitLengthVectors(50)
    a = SemanticPointer(next(gen), vocab=Vocabulary(50))  # noqa: F841
    b = SemanticPointer(next(gen), vocab=Vocabulary(50))  # noqa: F841
    with pytest.raises(SpaTypeError):
        eval(op)


@pytest.mark.parametrize("op", ("a+b", "a-b", "a*b", "a.dot(b)", "a.compare(b)"))
def test_none_vocab_is_always_compatible(op):
    gen = UnitLengthVectors(50)
    v = Vocabulary(50)
    a = SemanticPointer(next(gen), vocab=v)  # noqa: F841
    b = SemanticPointer(next(gen), vocab=None)  # noqa: F841
    eval(op)  # no assertion, just checking that no exception is raised


def test_fixed_pointer_network_creation():
    with spa.Network():
        A = SemanticPointer(next(UnitLengthVectors(16)))
        node = A.construct()
    assert_equal(node.output, A.v)


@pytest.mark.parametrize("op", ["+", "-", "*"])
@pytest.mark.parametrize("order", ["AB", "BA"])
def test_binary_operation_on_fixed_pointer_with_pointer_symbol(
    Simulator, op, order, seed, rng
):
    vocab = spa.Vocabulary(64, pointer_gen=rng)
    vocab.populate("A; B")
    a = PointerSymbol("A", TVocabulary(vocab))  # noqa: F841
    b = SemanticPointer(vocab["B"].v)  # noqa: F841

    with spa.Network(seed=seed) as model:
        if order == "AB":
            x = eval("a" + op + "b")
        elif order == "BA":
            x = eval("b" + op + "a")
        else:
            raise ValueError("Invalid order argument.")
        p = nengo.Probe(x.construct(), synapse=0.03)

    with Simulator(model) as sim:
        sim.run(0.5)

    assert_sp_close(
        sim.trange(), sim.data[p], vocab.parse(order[0] + op + order[1]), skip=0.3
    )


@pytest.mark.parametrize("op", ("+", "*"))
def test_incompatible_algebra(op):
    gen = UnitLengthVectors(32)
    a = SemanticPointer(next(gen), algebra=AbstractAlgebra())  # noqa: F841
    b = SemanticPointer(next(gen), algebra=AbstractAlgebra())  # noqa: F841
    with pytest.raises(TypeError):
        eval("a" + op + "b")


def test_invalid_algebra():
    gen = UnitLengthVectors(32)
    with pytest.raises(ValidationError, match="AbstractAlgebra"):
        SemanticPointer(next(gen), algebra=HrrAlgebra)
    SemanticPointer(next(gen), algebra=HrrAlgebra())


def test_identity(algebra):
    assert np.allclose(Identity(64, algebra=algebra).v, algebra.identity_element(64))


def test_absorbing_element(algebra, plt):
    plt.plot([0, 1], [0, 1])
    try:
        assert np.allclose(
            AbsorbingElement(64, algebra=algebra).v, algebra.absorbing_element(64)
        )
    except NotImplementedError:
        pass


def test_zero(algebra):
    assert np.allclose(Zero(64, algebra=algebra).v, algebra.zero_element(64))


def test_name():
    a = SemanticPointer(np.ones(4), name="a")
    b = SemanticPointer(np.ones(4), name="b")
    unnamed = SemanticPointer(np.ones(4), name=None)

    assert str(a) == "SemanticPointer<a>"
    assert repr(a) == (
        "SemanticPointer({!r}, vocab={!r}, algebra={!r}, name={!r}".format(
            a.v, a.vocab, a.algebra, a.name
        )
    )

    assert (-a).name == "-(a)"
    assert (~a).name == "~(a)"
    assert a.normalized().name == "(a).normalized()"
    assert a.unitary().name == "(a).unitary()"
    assert (a + b).name == "(a)+(b)"
    assert (a * b).name == "(a)*(b)"
    assert (2.0 * a).name == "(2.0)*(a)"
    assert (a / 2.0).name == "(a)/(2.0)"

    assert (a + unnamed).name is None
    assert (a * unnamed).name is None

    # check that names that blow up exponentially in length are truncated
    for i in range(10):
        a += a * b
    assert len(a.name) == a.MAX_NAME
    assert a.name.endswith("...")


def test_translate(rng):
    v1 = spa.Vocabulary(16, pointer_gen=rng)
    v1.populate("A; B")
    v2 = spa.Vocabulary(16, pointer_gen=rng)
    v2.populate("A; B")
    v3 = spa.Vocabulary(16, pointer_gen=rng)
    v3.populate("B")

    a1 = v1.parse("A")
    assert isinstance(a1, SemanticPointer)

    assert np.allclose(a1.translate(v2).dot(v2["A"]), 1.0, atol=0.2)
    assert np.allclose(a1.translate(v2, keys=["A"]).dot(v2["A"]), 1.0, atol=0.2)
    assert not np.allclose(a1.translate(v2, keys=["B"]).dot(v2["A"]), 1.0, atol=0.2)

    with pytest.warns(NengoWarning, match="source vocabulary has keys not existent"):
        a1.translate(v3)
    assert np.allclose(a1.translate(v3, populate=True).dot(v3["A"]), 1.0, atol=0.2)



import nengo
import numpy as np


class MatrixMult(nengo.Network):
    """Computes the matrix product A*B.

    Both matrices need to be two dimensional.

    See the `Matrix Multiplication example <:doc:networks>`_
    for a description of the network internals.

    Parameters
    ----------
    n_neurons : int
        Number of neurons used per product of two scalars.

        .. note:: If an odd number of neurons is given, one less neuron will be
                  used per product to obtain an even number. This is due to
                  the implementation the `.Product` network.
    shape_left : tuple
        Shape of the A input matrix.
    shape_right : tuple
        Shape of the B input matrix.
    **kwargs : dict
        Keyword arguments to pass through to the `nengo.Network` constructor.

    Attributes
    ----------
    input_left : nengo.Node
        The left matrix (A) to multiply.
    input_right : nengo.Node
        The left matrix (A) to multiply.
    C : nengo.networks.Product
        The product network doing the matrix multiplication.
    output : nengo.node
        The resulting matrix result.
    """

    def __init__(self, n_neurons, shape_left, shape_right, **kwargs):
        if len(shape_left) != 2:
            raise ValueError("Shape {} is not two dimensional.".format(shape_left))
        if len(shape_right) != 2:
            raise ValueError("Shape {} is not two dimensional.".format(shape_right))
        if shape_left[1] != shape_right[0]:
            raise ValueError(
                "Matrix dimensions {} and  {} are incompatible".format(
                    shape_left, shape_right
                )
            )

        super().__init__(**kwargs)

        size_left = np.prod(shape_left)
        size_right = np.prod(shape_right)

        with self:
            self.input_left = nengo.Node(size_in=size_left)
            self.input_right = nengo.Node(size_in=size_right)

            # The C matrix is composed of populations that each contain
            # one element of A (left) and one element of B (right).
            # These elements will be multiplied together in the next step.
            size_c = size_left * shape_right[1]
            self.C = nengo.networks.Product(n_neurons, size_c)

            # Determine the transformation matrices to get the correct pairwise
            # products computed.  This looks a bit like black magic but if
            # you manually try multiplying two matrices together, you can see
            # the underlying pattern.  Basically, we need to build up D1*D2*D3
            # pairs of numbers in C to compute the product of.  If i,j,k are
            # the indexes into the D1*D2*D3 products, we want to compute the
            # product # of element (i,j) in A with the element (j,k) in B.  The
            # index in # A of (i,j) is j+i*D2 and the index in B of (j,k) is
            # k+j*D3. The index in C is j+k*D2+i*D2*D3, multiplied by 2 since
            # there are # two values per ensemble.  We add 1 to the B index so
            # it goes into # the second value in the ensemble.
            transform_left = np.zeros((size_c, size_left))
            transform_right = np.zeros((size_c, size_right))

            for i, j, k in np.ndindex(shape_left[0], *shape_right):
                c_index = j + k * shape_right[0] + i * size_right
                transform_left[c_index][j + i * shape_right[0]] = 1
                transform_right[c_index][k + j * shape_right[1]] = 1

            nengo.Connection(
                self.input_left, self.C.input_a, transform=transform_left, synapse=None
            )
            nengo.Connection(
                self.input_right,
                self.C.input_b,
                transform=transform_right,
                synapse=None,
            )

            # Now do the appropriate summing
            size_output = shape_left[0] * shape_right[1]
            self.output = nengo.Node(size_in=size_output)

            # The mapping for this transformation is much easier, since we want
            # to combine D2 pairs of elements (we sum D2 products together)
            transform_c = np.zeros((size_output, size_c))
            for i in range(size_c):
                transform_c[i // shape_right[0]][i] = 1

            nengo.Connection(
                self.C.output, self.output, transform=transform_c, synapse=None
            )



from nengo.params import BoolParam, Default, IntParam

from nengo_spa.network import Network
from nengo_spa.vocabulary import VocabularyOrDimParam


class Bind(Network):
    """Network for binding together two inputs.

    Parameters
    ----------
    vocab : Vocabulary or int
        The vocabulary to use to interpret the vector. If an integer is given,
        the default vocabulary of that dimensionality will be used.
    neurons_per_dimension : int, optional (Default: 200)
        Number of neurons to use in each dimension.
    unbind_left : bool, optional
        Whether to unbind the left input.
    unbind_right : bool, optional
        Whether to unbind the right input.
    **kwargs : dict
        Keyword arguments passed through to `nengo_spa.Network`.

    Attributes
    ----------
    input_left : nengo.Node
        Left input vector.
    input_right : nengo.Node
        Right input vector.
    output : nengo.Node
        Output.
    """

    vocab = VocabularyOrDimParam("vocab", default=None, readonly=True)
    neurons_per_dimension = IntParam(
        "neurons_per_dimension", default=200, low=1, readonly=True
    )
    unbind_left = BoolParam("unbind_left", default=False, readonly=True)
    unbind_right = BoolParam("unbind_right", default=False, readonly=True)

    def __init__(
        self,
        vocab=Default,
        neurons_per_dimension=Default,
        unbind_left=Default,
        unbind_right=Default,
        **kwargs
    ):
        super(Bind, self).__init__(**kwargs)

        self.vocab = vocab
        self.neurons_per_dimension = neurons_per_dimension
        self.unbind_left = unbind_left
        self.unbind_right = unbind_right

        with self:
            self.binding_net, inputs, output = self.vocab.algebra.implement_binding(
                self.neurons_per_dimension,
                self.vocab.dimensions,
                self.unbind_left,
                self.unbind_right,
            )

        self.input_left = inputs[0]
        self.input_right = inputs[1]
        self.output = output

        self.declare_input(self.input_left, self.vocab)
        self.declare_input(self.input_right, self.vocab)
        self.declare_output(self.output, self.vocab)



from abc import ABCMeta
import warnings


class _DuckTypedABCMeta(ABCMeta):
    def __instancecheck__(cls, instance):
        if super().__instancecheck__(instance):
            return True

        for member in dir(cls):
            if member.startswith("_"):
                continue
            if not hasattr(instance, member) or not hasattr(
                getattr(instance, member), "__self__"
            ):
                return False
        warnings.warn(
            DeprecationWarning(
                "Please do not rely on pure duck-typing for {clsname}. "
                "Explicitly register your class {userclass} as a virtual subclass "
                "of {clsname} or derive from it.".format(
                    clsname=cls.__name__, userclass=instance.__class__.__name__
                )
            )
        )
        return True


class AbstractAlgebra(metaclass=_DuckTypedABCMeta):
    """Abstract base class for algebras.

    Custom algebras can be defined by implementing the interface of this
    abstract base class.
    """

    def is_valid_dimensionality(self, d):
        """Checks whether *d* is a valid vector dimensionality.

        Parameters
        ----------
        d : int
            Dimensionality

        Returns
        -------
        bool
            *True*, if *d* is a valid vector dimensionality for the use with
            the algebra.
        """
        raise NotImplementedError()

    def make_unitary(self, v):
        """Returns a unitary vector based on the vector *v*.

        A unitary vector does not change the length of a vector it is bound to.

        Parameters
        ----------
        v : (d,) ndarray
            Vector to base unitary vector on.

        Returns
        -------
        ndarray
            Unitary vector.
        """
        raise NotImplementedError()

    def superpose(self, a, b):
        """Returns the superposition of *a* and *b*.

        This is commonly elementwise addition.

        Parameters
        ----------
        a : (d,) ndarray
            Left operand in superposition.
        b : (d,) ndarray
            Right operand in superposition.

        Returns
        -------
        (d,) ndarray
            Superposed vector.
        """
        raise NotImplementedError()

    def bind(self, a, b):
        """Returns the binding of *a* and *b*.

        The resulting vector should in most cases be dissimilar to both inputs.

        Parameters
        ----------
        a : (d,) ndarray
            Left operand in binding.
        b : (d,) ndarray
            Right operand in binding.

        Returns
        -------
        (d,) ndarray
            Bound vector.
        """
        raise NotImplementedError()

    def invert(self, v):
        """Invert vector *v*.

        A vector bound to its inverse will result in the identity vector.

        Parameters
        ----------
        v : (d,) ndarray
            Vector to invert.

        Returns
        -------
        (d,) ndarray
            Inverted vector.
        """
        raise NotImplementedError()

    def get_binding_matrix(self, v, swap_inputs=False):
        """Returns the transformation matrix for binding with a fixed vector.

        Parameters
        ----------
        v : (d,) ndarray
            Fixed vector to derive binding matrix for.
        swap_inputs : bool, optional
            By default the matrix will be such that *v* becomes the *right*
            operand in the binding. By setting *swap_inputs*, the matrix will
            be such that *v* becomes the *left* operand. For binding operations
            that are commutative (such as circular convolution), this has no
            effect.

        Returns
        -------
        (d, d) ndarray
            Transformation matrix to perform binding with *v*.
        """
        raise NotImplementedError()

    def get_inversion_matrix(self, d):
        """Returns the transformation matrix for inverting a vector.

        Parameters
        ----------
        d : int
            Vector dimensionality (determines the matrix size).

        Returns
        -------
        (d, d) ndarray
            Transformation matrix to invert a vector.
        """
        raise NotImplementedError()

    def implement_superposition(self, n_neurons_per_d, d, n):
        """Implement neural network for superposing vectors.

        Parameters
        ----------
        n_neurons_per_d : int
            Neurons to use per dimension.
        d : int
            Dimensionality of the vectors.
        n : int
            Number of vectors to superpose in the network.

        Returns
        -------
        tuple
            Tuple *(net, inputs, output)* where *net* is the implemented
            `nengo.Network`, *inputs* a sequence of length *n* of inputs to the
            network, and *output* the network output.
        """
        raise NotImplementedError()

    def implement_binding(self, n_neurons_per_d, d, unbind_left, unbind_right):
        """Implement neural network for binding vectors.

        Parameters
        ----------
        n_neurons_per_d : int
            Neurons to use per dimension.
        d : int
            Dimensionality of the vectors.
        unbind_left : bool
            Whether the left input should be unbound from the right input.
        unbind_right : bool
            Whether the right input should be unbound from the left input.

        Returns
        -------
        tuple
            Tuple *(net, inputs, output)* where *net* is the implemented
            `nengo.Network`, *inputs* a sequence of the left and the right
            input in that order, and *output* the network output.
        """
        raise NotImplementedError()

    def absorbing_element(self, d):
        """Return the standard absorbing element of dimensionality *d*.

        An absorbing element will produce a scaled version of itself when bound
        to another vector. The standard absorbing element is the absorbing
        element with norm 1.

        Some algebras might not have an absorbing element other than the zero
        vector. In that case a *NotImplementedError* may be raised.

        Parameters
        ----------
        d : int
            Vector dimensionality.

        Returns
        -------
        (d,) ndarray
            Standard absorbing element.
        """
        raise NotImplementedError()

    def identity_element(self, d):
        """Return the identity element of dimensionality *d*.

        The identity does not change the vector it is bound to.

        Parameters
        ----------
        d : int
            Vector dimensionality.

        Returns
        -------
        (d,) ndarray
            Identity element.
        """
        raise NotImplementedError()

    def zero_element(self, d):
        """Return the zero element of dimensionality *d*.

        The zero element produces itself when bound to a different vector.
        Usually this will be the zero vector.

        Parameters
        ----------
        d : int
            Vector dimensionality.

        Returns
        -------
        (d,) ndarray
            Zero element.
        """
        raise NotImplementedError()



"""NengoSPA version information.

We use semantic versioning (see https://semver.org/).
and conform to PEP440 (see https://www.python.org/dev/peps/pep-0440/).
'.devN' will be added to the version unless the code base represents
a release version. Release versions are git tagged with the version.
"""

name = "nengo_spa"
version_info = (1, 1, 1)  # (major, minor, patch)
dev = None

version = "{v}{dev}".format(
    v=".".join(str(v) for v in version_info),
    dev="" if dev is None else ".dev{:d}".format(dev),
)



import pytest

import nengo_spa as spa
from nengo_spa.exceptions import SpaParseError, SpaTypeError


def test_missing_source():
    d = 16
    with spa.Network():
        a = spa.State(d)
        assert a
        with pytest.raises(NameError):
            b >> a  # noqa: F821


@pytest.mark.parametrize("sink", ("b", "B"))
def test_missing_sink(sink):
    d = 16
    with spa.Network():
        a = spa.State(d)
        assert a
        with pytest.raises(NameError):
            eval("a >> %s" % sink)


@pytest.mark.parametrize("sink", ("0.5", "0.6 * a", "a * 0.6"))
def test_invalid_sink(sink):
    d = 16
    with spa.Network():
        a = spa.State(d)
        assert a
        with pytest.raises((SyntaxError, SpaTypeError)):
            eval("a >> {}".format(sink))


def test_missing_pointer():
    vocab = spa.Vocabulary(16)
    with spa.Network():
        a = spa.State(vocab)
        assert a
        with pytest.raises(SpaParseError):
            spa.sym.A >> a



import nengo
from nengo.dists import CosineSimilarity
from nengo.exceptions import ValidationError
import numpy as np

from nengo_spa.networks.matrix_multiplication import MatrixMult


def calc_sub_d(dimensions):
    sub_d = int(np.sqrt(dimensions))
    if sub_d * sub_d != dimensions:
        raise ValidationError("Dimensions must be a square number.", "dimensions")
    return sub_d


def inversion_matrix(dimensions):
    sub_d = calc_sub_d(dimensions)
    m = np.zeros((dimensions, dimensions))
    for i in range(dimensions):
        j = sub_d * i
        m[j % dimensions + j // dimensions, i] = 1.0
    return m


def swapping_matrix(dimensions):
    sub_d = calc_sub_d(dimensions)
    m = np.zeros((dimensions, dimensions))
    for i in range(dimensions):
        j = i // sub_d + sub_d * (i % sub_d)
        m[i, j] = 1.0
    return m


class VTB(nengo.Network):
    r"""Compute vector-derived transformation binding (VTB).

    VTB uses elementwise addition for superposition. The binding operation
    :math:`\mathcal{B}(x, y)` is defined as

    .. math::

       \mathcal{B}(x, y) := V_y x = \left[\begin{array}{ccc}
           V_y' &    0 &    0 \\
              0 & V_y' &    0 \\
              0 &    0 & V_y'
           \end{array}\right] x

    with

    .. math::

       V_y' = d^{\frac{1}{4}} \left[\begin{array}{cccc}
           y_1            & y_2            & \dots  & y_{d'}  \\
           y_{d' + 1}     & y_{d' + 2}     & \dots  & y_{2d'} \\
           \vdots         & \vdots         & \ddots & \vdots  \\
           y_{d - d' + 1} & y_{d - d' + 2} & \dots  & y_d
       \end{array}\right]

    and

    .. math:: d'^2 = d.

    The approximate inverse :math:`y^+` for :math:`y` is permuting the elements
    such that :math:`V_{y^+} = V_y`.

    Note that VTB requires the vector dimensionality to be square.

    The VTB binding operation is neither associative nor commutative.

    Publications with further information are forthcoming.

    Parameters
    ----------
    n_neurons : int
        Number of neurons to use in each product computation.
    dimensions : int
        The number of dimensions of the input and output vectors. Needs to be a
        square number.
    unbind_left : bool
        Whether to unbind the left input vector from the right input vector.
    unbind_right : bool
        Whether to unbind the right input vector from the left input vector.
    **kwargs : dict
        Keyword arguments to pass through to the `nengo.Network` constructor.

    Attributes
    ----------
    input_left : nengo.Node
        The left operand vector to be bound.
    input_right : nengo.Node
        The right operand vector to be bound.
    mat : nengo.Node
        Representation of the matrix :math:`V_y'`.
    vec : nengo.Node
        Representation of the vector :math:`y`.
    matmuls : list
        Matrix multiplication networks.
    output : nengo.Node
        The resulting bound vector.
    """

    def __init__(
        self, n_neurons, dimensions, unbind_left=False, unbind_right=False, **kwargs
    ):
        super().__init__(**kwargs)

        sub_d = calc_sub_d(dimensions)

        shape_left = (sub_d, sub_d)
        shape_right = (sub_d, 1)

        with self:
            self.input_left = nengo.Node(size_in=dimensions)
            self.input_right = nengo.Node(size_in=dimensions)
            self.output = nengo.Node(size_in=dimensions)

            self.mat = nengo.Node(size_in=dimensions)
            self.vec = nengo.Node(size_in=dimensions)

            if unbind_left and unbind_right:
                raise ValueError("Cannot unbind both sides at the same time.")
            elif unbind_left:
                nengo.Connection(
                    self.input_left,
                    self.mat,
                    transform=inversion_matrix(dimensions),
                    synapse=None,
                )
                nengo.Connection(
                    self.input_right,
                    self.vec,
                    transform=swapping_matrix(dimensions),
                    synapse=None,
                )
            else:
                nengo.Connection(self.input_left, self.vec, synapse=None)
                if unbind_right:
                    tr = inversion_matrix(dimensions)
                else:
                    tr = 1.0
                nengo.Connection(self.input_right, self.mat, transform=tr, synapse=None)

            with nengo.Config(nengo.Ensemble) as cfg:
                cfg[nengo.Ensemble].intercepts = CosineSimilarity(dimensions + 2)
                cfg[nengo.Ensemble].eval_points = CosineSimilarity(dimensions + 2)
                self.matmuls = [
                    MatrixMult(n_neurons, shape_left, shape_right) for i in range(sub_d)
                ]

            for i in range(sub_d):
                mm = self.matmuls[i]
                sl = slice(i * sub_d, (i + 1) * sub_d)
                nengo.Connection(self.mat, mm.input_left, synapse=None)
                nengo.Connection(self.vec[sl], mm.input_right, synapse=None)
                nengo.Connection(
                    mm.output, self.output[sl], transform=np.sqrt(sub_d), synapse=None
                )



import inspect
import weakref

import nengo
from nengo.config import Config, SupportDefaultsMixin
import numpy as np

from nengo_spa.action_selection import ifmax as actions_ifmax
from nengo_spa.ast.base import Noop
from nengo_spa.connectors import (
    as_ast_node,
    input_vocab_registry,
    output_vocab_registry,
    SpaOperatorMixin,
)
from nengo_spa.types import TScalar
from nengo_spa.vocabulary import VocabularyMap, VocabularyMapParam


class _AutoConfig:
    def __init__(self, cfg):
        self._cfg = cfg

    def __getattr__(self, name):
        return getattr(self._cfg, name)

    def __getitem__(self, key):
        if inspect.isclass(key) and key not in self._cfg.params:
            self._cfg.configures(key)
        return self._cfg[key]


def ifmax(name, condition=None, *actions):
    """Defines a potential action within an `ActionSelection` context.

    This implementation allows Nengo objects in addition to AST nodes as
    condition argument.

    Parameters
    ----------
    name : str, optional
        Name for the action. Can be omitted.
    condition : nengo_spa.ast.base.Node or NengoObject
        The utility value for the given actions.
    actions : sequence of `RoutedConnection`
        The actions to activate if the given utility is the highest.

    Returns
    -------
    NengoObject
        Nengo object that can be connected to, to provide additional input to
        the utility value. It is possible (but not necessary) to use SPA style
        connections of the form ``scalar >> utility`` to this object.
    """
    if not isinstance(name, str):
        if condition is not None:
            actions = (condition,) + actions
        condition = name
        name = None

    if condition is None:
        raise ValueError("Must provide `condition` (though it may be 0).")
    elif condition == 0:
        condition = Noop(TScalar)
    else:
        condition = as_ast_node(condition)

    return actions_ifmax(name, as_ast_node(condition), *actions)


class Network(nengo.Network, SupportDefaultsMixin, SpaOperatorMixin):
    """Base class for SPA networks or modules.

    SPA modules are networks that declare their inputs and outputs with
    associated `.Vocabulary` instances. These inputs and outputs can then be
    be used in the SPA syntax, for example ``module1.output >> module2.input``.
    Inputs and outputs named `default` can be omitted in the SPA syntax so that
    one can write ``module1 >> module2``.

    Furthermore, SPA modules allow to configure parameters of contained SPA
    modules, for example::

        with spa.Network() as net:
            net.config[spa.State].vocab = 32
            state = spa.State()  # Will now have a 32-dimensional vocabulary

    Parameters
    ----------
    label : str, optional
        Name of the network.
    seed : int, optional
        Random number seed for the network.
    add_to_container : bool, optional
        Determines if this network will be added to the current container.
    vocabs : VocabularyMap, optional
        Maps from integer dimensionalities to the associated default
        vocabularies.

    Attributes
    ----------
    vocabs : VocabularyMap
        Maps from integer dimensionalities to the associated default
        vocabularies.
    """

    _master_vocabs = weakref.WeakKeyDictionary()
    vocabs = VocabularyMapParam("vocabs", default=None, optional=False)

    _input_types = {}
    _output_types = {}

    def __init__(self, label=None, seed=None, add_to_container=None, vocabs=None):
        super(Network, self).__init__(label, seed, add_to_container)
        self.config.configures(Network)

        if vocabs is None:
            vocabs = Config.default(Network, "vocabs")
            if vocabs is None and len(Network.context) > 0:
                vocabs = self._master_vocabs.get(Network.context[0], None)
            if vocabs is None:
                if seed is not None:
                    rng = np.random.RandomState(seed)
                else:
                    rng = None
                vocabs = VocabularyMap(rng=rng)
                if len(Network.context) > 0:
                    self.__class__._master_vocabs[Network.context[0]] = vocabs
        self.vocabs = vocabs
        self.config[Network].vocabs = vocabs

        self._stimuli = None

    @property
    def config(self):
        return _AutoConfig(self._config)

    @classmethod
    def get_input_vocab(cls, obj):
        """Get the vocabulary associated with an network input *obj*."""
        return input_vocab_registry[obj]

    @classmethod
    def get_output_vocab(cls, obj):
        """Get the vocabulary associated with an network output *obj*."""
        return output_vocab_registry[obj]

    def declare_input(self, obj, vocab):
        """Declares a network input.

        Parameters
        ----------
        obj : nengo.base.NengoObject
            Nengo object to use as an input to the network.
        vocab: Vocabulary
            Vocabulary to assign to the input.
        """
        return input_vocab_registry.declare_connector(obj, vocab)

    def declare_output(self, obj, vocab):
        """Declares a network output.

        Parameters
        ----------
        obj : nengo.base.NengoObject
            Nengo object to use as an output of the network.
        vocab : Vocabulary
            Vocabulary to assign to the output.
        """
        return output_vocab_registry.declare_connector(obj, vocab)


def create_inhibit_node(net, strength=2.0, **kwargs):
    """Creates a node that inhibits all ensembles in a network.

    Parameters
    ----------
    net : nengo.Network
        Network to inhibit.
    strength : float
        Strength of the inhibition.
    **kwargs : dict
        Additional keyword arguments for the created connections from the node
        to the inhibited ensemble neurons.

    Returns
    -------
    nengo.Node
        Node that can be connected to, to provide an inhibitory signal to the
        network.
    """
    inhibit_node = nengo.Node(size_in=1)
    for e in net.all_ensembles:
        nengo.Connection(
            inhibit_node,
            e.neurons,
            transform=-strength * np.ones((e.n_neurons, 1), **kwargs),
        )
    return inhibit_node
